from policyengine_us.model_api import *


class mi_homestead_property_tax_credit_countable_property_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan homestead property tax credit countable property tax (including rent equivalent)"
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-520",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040CR.pdf#page=1",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.credits.homestead_property_tax
        # Line 10
        property_tax = add(tax_unit, period, ["real_estate_taxes"])
        # Line 11
        rent = add(tax_unit, period, ["rent"])
        # Line 12
        applicable_rent = rent * p.rent_equivalization
        # Line 13
        return property_tax + applicable_rent



from policyengine_us.model_api import *


class foreign_tax_credit_credit_limit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Foreign tax credit credit limit"
    definition_period = YEAR
    documentation = "Foreign tax credit from Form 1116"
    unit = USD

    # The foreign tax credit is the first non-refundable credit in Schedule 3 (Form 1040)
    adds = ["income_tax_before_credits"]



from policyengine_us.model_api import *


class hi_dependent_care_benefits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii Dependent Care Benefits"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = (
        "https://files.hawaii.gov/tax/legal/hrs/hrs_235.pdf#page=40"
        "https://files.hawaii.gov/tax/forms/2022/schx_i.pdf#page=1"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=28"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.hi.tax.income.credits.cdcc.dependent_care_benefits
        p_irs = parameters(
            period
        ).gov.irs.gross_income.dependent_care_assistance_programs
        # Schedule X PART II:
        # line 2:
        dcb_amount = add(
            tax_unit, period, ["dependent_care_employer_benefits"]
        )
        # line 3, 4 are ignored, so line 5 = line 2
        # line 6:
        qualified_expense_amount = tax_unit(
            "tax_unit_childcare_expenses", period
        )
        # line 7 = min(line 5, line 6) = min(line 2, line 6):
        capped_expenses = min_(dcb_amount, qualified_expense_amount)
        # line 8 & line 9:
        # married persons must file a joint return to claim the credit
        # if single, the min will be his/her income
        min_head_spouse_earned = tax_unit(
            "hi_cdcc_min_head_spouse_earned", period
        )
        # line 10 = min(line 7, line 8, line 9):
        earned_income_cap = min_(capped_expenses, min_head_spouse_earned)
        # line 11:
        filing_status = tax_unit("filing_status", period)
        dcb_baseline = p_irs.reduction_amount[filing_status]
        # line 12 is ignored, since we do not consider sole proprietorship/partnership
        # line 13 is ignored, since it requires line 12
        # line 14:
        deductible_benefit = min_(earned_income_cap, dcb_baseline)
        # line 15 excluded_benefit = 0 since we ignore line 12
        # line 16 Taxable Benefit is ignored, since we never use it
        # line 17:
        expenses_amount = p.expense_floor.calc(
            tax_unit("count_cdcc_eligible", period)
        )
        # line 18 = line 14 + line 15 = line 14
        # line 19 = line 17 - line 18 = line 17 - line 14
        net_expenses = max_(0, expenses_amount - deductible_benefit)
        # line 20 = line 21 (d) = line 6
        # return line 22 = min(line 19, line 20) since we complete section B
        return min_(net_expenses, qualified_expense_amount)



from policyengine_us.model_api import *


class estate_tax_before_credits(Variable):
    value_type = float
    entity = Person
    label = "Estate tax before credits"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/2001#b_1"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.tax.estate.rate
        taxable_estate_value = person("taxable_estate_value", period)
        return p.calc(taxable_estate_value)



from policyengine_us.model_api import *


class va_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
        "https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2022/760-2022.pdf#page=1"
    )
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("va_agi", period)
        ded = tax_unit("va_deductions", period)
        exemptions = tax_unit("va_total_exemptions", period)
        total_deductions = ded + exemptions
        # Virginia allows a deduction (atop itemized deductions) for the
        # relevant expenses for the federal Child and Dependent Care Credit.
        cdcc_expense_deduction = tax_unit(
            "va_child_dependent_care_expense_deduction", period
        )
        return max_(agi - total_deductions - cdcc_expense_deduction, 0)



from policyengine_us.model_api import *


class person_in_poverty(Variable):
    label = "person in poverty"
    documentation = "Whether person is in poverty"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person.spm_unit("in_poverty", period)



from policyengine_us.model_api import *


class mi_is_senior_for_tax(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Michigan filer is a senior"
    definition_period = YEAR
    reference = "http://legislature.mi.gov/doc.aspx?mcl-206-514"
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mi.tax.income
        return tax_unit("greater_age_head_spouse", period) >= p.senior_age



from policyengine_us.model_api import *


class ny_college_tuition_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York itemized deduction for college tuition expenses"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.nysenate.gov/legislation/laws/TAX/615",  # (d)(4)
        "https://www.nysenate.gov/legislation/laws/TAX/606",  # (t)
    )
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ny.tax.income.deductions.itemized.college_tuition
        allowable_expenses = tax_unit(
            "ny_allowable_college_tuition_expenses", period
        )
        return allowable_expenses * p.applicable_percentage



from policyengine_us.model_api import *


class ok_federal_eitc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for federal EITC for the Oklahoma EITC computation"
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/oklahoma/title-68/section-68-2357-43/"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        eitc = parameters.gov.irs.credits.eitc(f"2020-01-01")
        investment_income_eligible = tax_unit(
            "ok_federal_eitc_investment_income_eligible", period
        )
        demographic_eligible = tax_unit(
            "ok_federal_eitc_demographic_eligible", period
        )
        # Define eligibility before considering separate filer limitation.
        eligible = demographic_eligible & investment_income_eligible
        # This parameter is true if separate filers are eligible.
        if eitc.eligibility.separate_filer:
            return eligible
        # If separate filers are not eligible, check if the filer is separate.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        return eligible & ~separate



from policyengine_us.model_api import *


class ctc_qualifying_child(Variable):
    value_type = bool
    entity = Person
    label = "CTC-qualifying child"
    documentation = "Child qualifies for the Child Tax Credit"
    definition_period = YEAR
    defined_for = "is_tax_unit_dependent"
    reference = "https://www.law.cornell.edu/uscode/text/26/24#c"

    def formula(person, period, parameters):
        age = person("age", period)
        p = parameters(period).gov.irs.credits.ctc
        age_limit = p.amount.base.thresholds[-1]
        age_eligible = age < age_limit
        meets_identification_requirements = person(
            "meets_ctc_child_identification_requirements", period
        )
        return age_eligible & meets_identification_requirements



from policyengine_us.model_api import *


class taxable_estate_value(Variable):
    value_type = float
    entity = Person
    label = "Taxable estate value"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/2001#b_1"



from policyengine_us.model_api import *


class or_tax_before_credits_in_prior_year(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR tax before credits in prior year"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR



from policyengine_us.model_api import *


class ne_social_security_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Nebraska social security subtraction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        fagi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.ne.tax.income.agi.subtractions
        oasdi_fraction = where(
            fagi <= p.social_security.threshold[filing_status],
            1.0,
            p.social_security.fraction,
        )
        taxable_oasdi = add(tax_unit, period, ["taxable_social_security"])
        return taxable_oasdi * oasdi_fraction



from policyengine_us.model_api import *


class va_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia refundable earned income tax credit"
    unit = USD
    documentation = "Refundable EITC credit reducing Virginia State income tax"
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=32"
    defined_for = "va_claims_refundable_eitc"

    adds = ["va_refundable_eitc_if_claimed"]



from policyengine_us.model_api import *


class ia_alternate_tax_consolidated(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa alternate tax for years on or after 2023"
    unit = USD
    definition_period = YEAR
    reference = "https://revenue.iowa.gov/media/2754/download?inline"
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        modified_income = tax_unit("ia_modified_income", period)
        # compute alternate tax following worksheet in the instructions
        p = parameters(period).gov.states.ia.tax.income.alternate_tax
        # ... determine alternate tax deduction
        elderly_head_or_spouse = (
            tax_unit("greater_age_head_spouse", period) >= p.elderly_age
        )
        alt_ded = where(
            elderly_head_or_spouse, p.deduction.elderly, p.deduction.nonelderly
        )
        # ... determine alternate tax amount
        alt_taxinc = max_(0, modified_income - alt_ded)
        return alt_taxinc * p.rate



from policyengine_us.model_api import *


class snap_individual_utility_allowance(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP Individual Utility Allowance"
    unit = USD
    documentation = "The individual utility allowance deduction for SNAP"
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        utility = parameters(period).gov.usda.snap.income.deductions.utility
        allowance_type = spm_unit("snap_utility_allowance_type", period)
        allowance_types = allowance_type.possible_values
        region = spm_unit.household("snap_utility_region_str", period)
        expense_types = utility.single.utility_types
        sum_of_individual_allowances = sum(
            [
                utility.single[expense.replace("_expense", "")][region]
                for expense in expense_types
            ]
        )
        return where(
            allowance_type == allowance_types.IUA,
            sum_of_individual_allowances,
            0,
        )



from policyengine_us.model_api import *


class employee_payroll_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "employee-side payroll tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_payroll_tax:
            return 0
        else:
            return add(
                person,
                period,
                [
                    "employee_social_security_tax",
                    "employee_medicare_tax",
                    "additional_medicare_tax",
                ],
            )



from policyengine_us.model_api import *


class age_head(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "Age of head of tax unit"
    documentation = "Age in years of taxpayer (i.e. primary adult)"
    unit = "year"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        age = person("age", period)
        head = person("is_tax_unit_head", period)
        return tax_unit.max(age * head)



from policyengine_us.model_api import *


class amt_lower_base_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax lower base tax amount"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) base tax, Form 6251 Part II Line 7 'All Others' - lower bracket"
    reference = "https://www.irs.gov/pub/irs-pdf/f6251.pdf"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.income.amt
        filing_status = tax_unit("filing_status", period)
        reduced_income = tax_unit("amt_income_less_exemptions", period)
        bracket_fraction = p.multiplier[filing_status]
        tax_rate_threshold = p.brackets.thresholds[-1] * bracket_fraction
        lower_rate = p.brackets.rates[0]
        return min_(reduced_income, tax_rate_threshold) * lower_rate



from policyengine_us.model_api import *
from policyengine_us.parameters.gov.hhs.medicaid.geography import (
    aca_rating_areas,
)


class slcsp_rating_area_default(Variable):
    value_type = int
    entity = Household
    label = (
        "Second-lowest ACA silver-plan cost rating area outside of LA County"
    )
    definition_period = YEAR

    def formula(household, period, parameters):
        # 1) Read the county string from the Household
        county_str = household("county_str", period)

        # 2) Ensure 'rating_area' is a string in aca_rating_areas
        aca_rating_areas["rating_area"] = aca_rating_areas[
            "rating_area"
        ].astype(str)

        # 3) Create a DataFrame using county_str
        df = pd.DataFrame({"county": county_str})

        # 4) Merge on "county"
        df_matched = pd.merge(
            df,
            aca_rating_areas,
            how="left",
            on="county",
        )

        # 5) Fill missing rating areas with "1", convert to int
        df_matched["rating_area"] = (
            df_matched["rating_area"].fillna("1").astype(int)
        )

        # 6) Return the rating_area column
        return df_matched["rating_area"]



from policyengine_us.model_api import *


class va_low_income_tax_credit_agi_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Virginia low income tax credit"
    definition_period = YEAR
    defined_for = "va_low_income_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        agi = tax_unit("va_agi", period)
        # Virginia does not account for blind or aged exemptions
        fpg = tax_unit("tax_unit_fpg", period)
        return agi <= fpg



from policyengine_us.model_api import *


class wa_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA
    adds = ["wa_income_tax_before_refundable_credits"]
    subtracts = ["wa_refundable_credits"]



from policyengine_us.model_api import *


class nj_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "New Jersey TANF resources eligible"
    definition_period = YEAR
    defined_for = StateCode.NJ

    def formula(spm_unit, period, parameters):
        # The WFNJ/TANF resource limit is $2,000 for an assistance unit.
        # https://www.nj.gov/humanservices/dfd/programs/workfirstnj/tanf_2021_23_st_plan.pdf#page=21
        p = parameters(period).gov.states.nj.njdhs.tanf.eligibility.resources
        countable_resources = spm_unit("nj_tanf_countable_resources", period)
        return countable_resources <= p.limit



from policyengine_us.model_api import *


class wi_homestead_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Wisconsin homestead credit eligibility status"
    definition_period = YEAR
    reference = "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2021/0013_homestead_tax_credit_informational_paper_13.pdf#page=7"
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wi.tax.income.credits
        # minimum age eligibility
        head_age = tax_unit("age_head", period)
        spouse_age = tax_unit("age_spouse", period)
        min_age = p.homestead.eligible.min_age
        age_eligible = (head_age >= min_age) | (spouse_age >= min_age)
        # earnings eligibility
        earnings = add(tax_unit, period, p.homestead.eligible.earnings_sources)
        has_positive_earnings = earnings > 0
        min_elderly_age = p.homestead.eligible.min_elderly_age
        head_is_elderly = head_age >= min_elderly_age
        spouse_is_elderly = spouse_age >= min_elderly_age
        is_elderly = head_is_elderly | spouse_is_elderly
        head_is_disabled = tax_unit("head_is_disabled", period)
        spouse_is_disabled = tax_unit("spouse_is_disabled", period)
        is_disabled = head_is_disabled | spouse_is_disabled
        earnings_eligible = has_positive_earnings | is_elderly | is_disabled
        # income eligibility
        homestead_income = tax_unit("wi_homestead_income", period)
        income_eligible = homestead_income < p.homestead.eligible.max_income
        # overall eligibility
        return age_eligible & earnings_eligible & income_eligible



from policyengine_us.model_api import *


class new_clean_vehicle_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "New clean vehicle credit"
    documentation = (
        "Nonrefundable credit for the purchase of a new clean vehicle"
    )
    unit = USD
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/30D",
        "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=373",
    )
    defined_for = "new_clean_vehicle_credit_eligible"

    def formula(tax_unit, period, parameters):
        credit_limit = tax_unit(
            "new_clean_vehicle_credit_credit_limit", period
        )
        potential = tax_unit("new_clean_vehicle_credit_potential", period)
        return min_(credit_limit, potential)



from policyengine_us.model_api import *


class sc_two_wage_earner_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina two wage earner credit"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/SC1040TT_2021.pdf",
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=23",
    )

    def formula(tax_unit, period, parameters):
        # Determine eligibility. Must be a joint filer.
        filing_status = tax_unit("filing_status", period)
        eligible = filing_status == filing_status.possible_values.JOINT
        # Find relevant gross income of head and spouse.
        person = tax_unit.members
        income = person("sc_gross_earned_income", period)
        head_income = tax_unit.max(income * person("is_tax_unit_head", period))
        spouse_income = tax_unit.max(
            income * person("is_tax_unit_spouse", period)
        )
        # Determine lesser of head and spouse income.
        lesser_head_spouse_income = min_(head_income, spouse_income)
        # Calculate credit based on rate.
        p = parameters(period).gov.states.sc.tax.income.credits.two_wage_earner
        credit_if_eligible = p.rate.calc(lesser_head_spouse_income)
        return credit_if_eligible * eligible



from policyengine_us.model_api import *


class taxable_ira_distributions(Variable):
    value_type = float
    entity = Person
    label = "Taxable IRA distributions"
    unit = USD
    documentation = "Taxable distributions from individual retirement accounts (typically traditional IRAs)."
    definition_period = YEAR



from policyengine_us.model_api import *


class md_married_or_has_child_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland non-refundable EITC for filers who are married or have qualifying child"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"  # (c)(1)
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)

        p = parameters(
            period
        ).gov.states.md.tax.income.credits.eitc.non_refundable.married_or_has_child
        # Limited to filers who are married or have child
        married_or_has_child = ~tax_unit(
            "md_qualifies_for_unmarried_childless_eitc", period
        )
        amount = federal_eitc * p.match
        return married_or_has_child * amount



from policyengine_us.model_api import *


class wa_working_families_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington Working Families Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://app.leg.wa.gov/RCW/default.aspx?cite=82.08.0206"
    defined_for = StateCode.WA

    def formula(tax_unit, period, parameters):
        # Filers must claim EITC and be in Washington to be eligible.
        # TODO: Include ITIN children.
        eligible = tax_unit("eitc", period) > 0
        # Parameters are based on EITC-eligible children.
        p = parameters(
            period
        ).gov.states.wa.tax.income.credits.working_families_tax_credit
        eitc_child_count = tax_unit("eitc_child_count", period)
        max_amount = p.amount.calc(eitc_child_count)
        # WFTC phases out at a certain amount below the EITC maximum AGI.
        # NB: The Revised Code of Washington is ambiguous:
        # "below the federal phase-out income"
        # The legislative analysis clarifies that this refers to "federal maximum AGI"
        # https://lawfilesext.leg.wa.gov/biennium/2021-22/Pdf/Bill%20Reports/House/1297-S.E%20HBR%20FBR%2021.pdf?q=20220706071752
        eitc_agi_limit = tax_unit("eitc_agi_limit", period)
        phase_out_start_reduction = p.phase_out.start_below_eitc.calc(
            eitc_child_count
        )
        phase_out_start = eitc_agi_limit - phase_out_start_reduction
        # The phase-out rates are hard-coded in the legal code, but HB 1888 (2021-22)
        # instructs DOR to revise it to get to zero by the EITC AGI limit.
        # https://app.leg.wa.gov/billsummary?BillNumber=1888&Year=2021&Initiative=false
        phase_out_rate = max_amount / phase_out_start_reduction
        earnings = tax_unit("filer_adjusted_earnings", period)
        excess = max_(0, earnings - phase_out_start)
        reduction = max_(0, excess * phase_out_rate)
        phased_out_amount = max_amount - reduction
        # minimum benefit applies if calculated amount exceeds zero
        amount_if_eligible = where(
            phased_out_amount > 0, max_(p.min_amount, phased_out_amount), 0
        )
        return amount_if_eligible * eligible



from policyengine_us.model_api import *


class ca_riv_general_relief_countable_vehicle_value(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County General Relief countable vehicle value"
    definition_period = YEAR
    defined_for = "in_riv"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.local.ca.riv.general_relief.property.vehicle_exemption
        total_vehicle_value = spm_unit.household(
            "household_vehicles_value", period
        )
        return max_(total_vehicle_value - p.amount, 0)



from policyengine_us.model_api import *


class EmployerPremiumContribution(Enum):
    NONE = "NONE"  # Employer paid none of premiums.
    SOME = "SOME"
    ALL = "ALL"
    NA = "N/A"


class employer_contribution_to_health_insurance_premiums_category(Variable):
    value_type = Enum
    entity = Person
    label = "Extent to which employer paid health insurance premiums"
    definition_period = YEAR
    possible_values = EmployerPremiumContribution
    default_value = EmployerPremiumContribution.NONE



from policyengine_us.model_api import *


class ma_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-2"
    defined_for = StateCode.MA

    adds = ["ma_part_a_agi", "ma_part_b_agi", "ma_part_c_agi"]



from policyengine_us.model_api import *


class commodity_supplemental_food_program(Variable):
    value_type = float
    entity = Person
    unit = USD
    definition_period = YEAR
    label = "Commodity Supplemental Food Program"
    defined_for = "commodity_supplemental_food_program_eligible"
    adds = ["gov.usda.csfp.amount"]



from policyengine_us.model_api import *


class va_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia non-refundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )
    defined_for = StateCode.VA

    adds = "gov.states.va.tax.income.credits.non_refundable"



from policyengine_us.model_api import *


class mt_old_age_subtraction(Variable):
    value_type = float
    entity = Person
    label = "Montana old age subtraction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.subtractions.old_age
        # Aged taxpayers are eligible for a subtraction amount
        age = person("age", period)
        return p.amount.calc(age)



from policyengine_us.model_api import *


class mn_cdcc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = (
        "Eligible for the Minnesota child and dependent care expense credit"
    )
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1cd_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1cd_22_0.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cdcc
        if p.separate_filers_excluded:
            filing_status = tax_unit("filing_status", period)
            return filing_status != filing_status.possible_values.SEPARATE
        else:
            return True



from policyengine_us.model_api import *


class az_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona total exemptions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    adds = [
        "az_aged_exemption",
        "az_blind_exemption",
        "az_parents_grandparents_exemption",
        "az_stillborn_exemption",
    ]



from policyengine_us.model_api import *


class il_total_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL total exemption allowance"
    unit = USD
    definition_period = YEAR

    defined_for = "il_is_exemption_eligible"

    adds = [
        "il_personal_exemption",
        "il_aged_blind_exemption",
        "il_dependent_exemption",
    ]



from policyengine_us.model_api import *


class hi_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii adjusted gross income"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    # Hawaii Instructions for Form N-11 Rev 2022 (p4)
    reference = " https://files.hawaii.gov/tax/forms/2022/n11ins.pdf"

    adds = ["hi_additions", "adjusted_gross_income"]
    subtracts = ["hi_subtractions"]
    # According to the tax form, Hawaii AGI can be negative.



from policyengine_us.model_api import *


class tx_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Texas Temporary Assistance for Needy Families (TANF)"
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-220-tanf",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-2420-eligibility-requirements",
        "https://www.law.cornell.edu/regulations/texas/title-1/part-15/chapter-372/subchapter-B",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        person = spm_unit.members

        # Must have at least one eligible child in certified group
        # (eligible child already includes immigration and SSI checks)
        has_eligible_child = spm_unit.any(
            person("tx_tanf_eligible_child", period)
        )

        # Financial eligibility
        income_eligible = spm_unit("tx_tanf_income_eligible", period)
        resources_eligible = spm_unit("tx_tanf_resources_eligible", period)

        return has_eligible_child & income_eligible & resources_eligible



from policyengine_us.model_api import *


class md_socsec_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD social security subtraction from AGI"
    unit = USD
    definition_period = YEAR
    reference = "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=14"
    defined_for = StateCode.MD
    adds = ["md_socsec_subtraction_amount"]



from policyengine_us.model_api import *


class de_taxable_income_joint(Variable):
    value_type = float
    entity = Person
    label = "Delaware taxable income when married filing jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        agi = person("de_agi_joint", period)
        is_head = person("is_tax_unit_head", period)
        total_agi = is_head * person.tax_unit.sum(agi)
        deductions = person("de_deduction_joint", period)
        total_deductions = person.tax_unit.sum(deductions)
        return max_(total_agi - total_deductions, 0)



from policyengine_us.model_api import *


class is_permanently_and_totally_disabled(Variable):
    value_type = bool
    entity = Person
    label = "Is permanently and totally disabled"
    definition_period = YEAR



from policyengine_us.model_api import *


class aca_ptc(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA premium tax credit for tax unit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/36B"
    defined_for = "is_aca_ptc_eligible"

    def formula(tax_unit, period, parameters):
        plan_cost = tax_unit("slcsp", period)
        income = tax_unit("aca_magi", period)
        applicable_figure = tax_unit("aca_ptc_phase_out_rate", period)
        takes_up_aca_if_eligible = tax_unit("takes_up_aca_if_eligible", period)
        return (
            max_(0, plan_cost - income * applicable_figure)
            * takes_up_aca_if_eligible
        )



from policyengine_us.model_api import *


class aca_magi_fraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA-related modified AGI as fraction of prior-year FPL"
    documentation = (
        "ACA-related MAGI as fraction of federal poverty line."
        "Documentation on use of prior-year FPL in the following reference:"
        "  title: 2022 IRS Form 8962 (ACA PTC) instructions, Line 4"
        "  href: https://www.irs.gov/pub/irs-pdf/i8962.pdf#page=7"
        "Documentation on truncation of fraction in the following reference:"
        "  title: 2022 IRS Form 8962 instructions, Line 5 Worksheet 2"
        "  href: https://www.irs.gov/pub/irs-pdf/i8962.pdf#page=8"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        magi = max_(0, tax_unit("aca_magi", period))
        fpg = tax_unit("tax_unit_fpg", period.last_year)
        return np.floor(100 * magi / fpg) / 100



from policyengine_us.model_api import *


class is_medically_needy_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Medically needy"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/42/part-435/subpart-D"

    def formula(person, period, parameters):
        in_category = person("is_in_medicaid_medically_needy_category", period)
        personal_income = person("ssi_countable_income", period)
        medical_expenses = person("medical_out_of_pocket_expenses", period)
        personal_assets = person("ssi_countable_resources", period)
        tax_unit = person.tax_unit
        income = tax_unit.sum(personal_income - medical_expenses)
        assets = tax_unit.sum(personal_assets)
        ma = parameters(period).gov.hhs.medicaid
        mn = ma.eligibility.categories.medically_needy
        is_joint = person.tax_unit("tax_unit_is_joint", period)
        state = person.household("state_code_str", period)
        income_limit = MONTHS_IN_YEAR * where(
            is_joint,
            mn.limit.income.couple[state],
            mn.limit.income.individual[state],
        )
        asset_limit = where(
            is_joint,
            mn.limit.assets.couple[state],
            mn.limit.assets.individual[state],
        )
        under_limits = (income <= income_limit) & (assets <= asset_limit)
        other_categories = [
            category
            for category in ma.eligibility.categories.covered
            if category != "is_medically_needy_for_medicaid"
        ]
        not_in_other_pathway = add(person, period, other_categories) == 0
        return in_category & not_in_other_pathway & under_limits



from policyengine_us.model_api import *


class id_2022_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho 2022 rebate"
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.id.tax.income.credits["2022_rebate"]
        income_tax_before_credits = tax_unit(
            "id_income_tax_before_non_refundable_credits", period
        )
        percentage_credit = p.rate * income_tax_before_credits
        floor = add(tax_unit, period, ["id_2022_rebate_floor"])
        return max_(percentage_credit, floor)



from policyengine_us.model_api import *


class ia_pension_exclusion_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Iowa pension exclusion"
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=26"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=26"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ia.tax.income.pension_exclusion
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        age = person("age", period)
        age_eligible = age >= p.minimum_age
        is_disabled = person("is_permanently_and_totally_disabled", period)
        return is_head_or_spouse & (age_eligible | is_disabled)



from policyengine_us.model_api import *


class hi_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii subtractions from federal adjusted gross income"
    reference = "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=13"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.subtractions
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class sc_retirement_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina retirement deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.SC

    adds = [
        "sc_retirement_deduction_indv",  # p1,p2
        "sc_retirement_deduction_survivors",  # p3
    ]



from policyengine_us.model_api import *


class mi_homestead_property_tax_credit_alternate_senior_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "Michigan alternate senior renter homestead property tax credit amount"
    )
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040CR.pdf#page=2",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=34",
    )
    defined_for = "mi_homestead_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.credits.homestead_property_tax
        senior = tax_unit("mi_is_senior_for_tax", period)
        # Line A
        allowable_credit_amount = tax_unit(
            "mi_homestead_property_tax_credit_pre_alternate_senior_amount",
            period,
        )
        # Line B
        rent = add(tax_unit, period, ["rent"])
        # Line C
        household_resources = tax_unit("mi_household_resources", period)
        applicable_resources = p.rate.senior.alternate * household_resources
        # Line D
        reduced_rent = max_(rent - applicable_resources, 0)
        senior_amount = max_(allowable_credit_amount, reduced_rent)
        capped_senior_amount = min_(senior_amount, p.cap)
        return senior * capped_senior_amount



from policyengine_us.model_api import *


class basic_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "basic income"
    unit = USD
    documentation = "Total basic income payments for this filer."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        eligible = tax_unit("basic_income_eligible", period)
        basic_income = tax_unit("basic_income_before_phase_out", period)
        phase_out = tax_unit("basic_income_phase_out", period)
        return eligible * (basic_income - phase_out)



from policyengine_us.model_api import *


class il_ccap_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible child for Illinois Child Care Assistance Program (CCAP)"
    definition_period = MONTH
    defined_for = StateCode.IL
    reference = "https://www.dhs.state.il.us/page.aspx?item=104995"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.ccap.age_limit
        age = person("monthly_age", period)
        is_disabled = person("is_disabled", period)
        age_limit = where(is_disabled, p.special_needs_child, p.child)
        age_eligible = age < age_limit
        is_dependent = person("is_tax_unit_dependent", period)
        immigration_status_eligible = person(
            "il_ccap_immigration_status_eligible_person", period
        )

        return age_eligible & is_dependent & immigration_status_eligible



from policyengine_us.model_api import *


class wi_retirement_income_subtraction_agi_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Wisconsin retirement income subtraction AGI eligibility"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB-inst.pdf#page=9"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleSBf.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleSB-Inst.pdf#page=7"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        fstatus = tax_unit("filing_status", period)
        p = parameters(period).gov.states.wi.tax.income
        return agi < p.subtractions.retirement_income.max_agi[fstatus]



from policyengine_us.model_api import *


class social_security(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Social Security benefits, not including SSI"
    label = "Social Security"
    unit = USD
    adds = [
        "social_security_" + i
        for i in ["dependents", "disability", "retirement", "survivors"]
    ]
    uprating = "calibration.gov.irs.soi.social_security"



from policyengine_us.model_api import *


class dc_disabled_exclusion_subtraction(Variable):
    value_type = float
    entity = Person
    label = "DC disabled exclusion subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=63"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=55"
        # More details on page 56.
        "https://code.dccouncil.gov/us/dc/council/code/titles/47/chapters/18/subchapters/III"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        # Determine disablity-related eligibility
        is_disabled = person("is_permanently_and_totally_disabled", period)
        gets_ssi_or_ssdi = (
            add(person, period, ["ssi", "social_security_disability"]) > 0
        )
        disabled_eligible = is_disabled & gets_ssi_or_ssdi
        # Determine income-related eligibility
        p = parameters(
            period
        ).gov.states.dc.tax.income.subtractions.disabled_exclusion
        # Program counts AGI of all persons residing in a household, except
        # those under a written lease.
        # Assume tax unit for no
        tax_unit = person.tax_unit
        agi = tax_unit("adjusted_gross_income", period)
        income_eligible = agi < p.income_limit
        # Return subtraction amount if meet both eligibility requirements.
        # Use total household income, even that excluded from federal AGI.
        # Limit to IRS countable income as inclusion of benefits are unclear.
        household_gross_income = add(
            person.household, period, ["irs_gross_income"]
        )
        capped_amount = min_(household_gross_income, p.amount)
        return disabled_eligible * income_eligible * capped_amount



from policyengine_us.model_api import *


class tx_harris_rides_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Harris County RIDES eligibility"
    reference = "https://rides.harriscountytx.gov/Registration/Registration-Information"
    defined_for = "in_harris_county_tx"

    def formula(person, period, parameters):
        p = parameters(period).gov.local.tx.harris.rides

        # Age eligibility
        age = person("age", period)
        age_eligible = age >= p.age_threshold

        # Disability eligibility
        has_disability = person("is_disabled", period)

        return age_eligible | has_disability



from policyengine_us.model_api import *


class utility_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Utility expenses"
    unit = USD
    definition_period = YEAR

    adds = [
        "heating_cooling_expense",
        "gas_expense",
        "electricity_expense",
        "trash_expense",
        "water_expense",
        "sewage_expense",
        "phone_expense",
    ]



from policyengine_us.model_api import *


class acp(Variable):
    value_type = float
    entity = SPMUnit
    label = "Affordable Connectivity Program"
    documentation = "Affordable Connectivity Program amount"
    definition_period = YEAR
    unit = USD
    defined_for = "is_acp_eligible"
    reference = "https://uscode.house.gov/view.xhtml?req=granuleid:USC-prelim-title47-section1752&edition=prelim"

    def formula(spm_unit, period, parameters):
        broadband_cost = spm_unit("broadband_cost_after_lifeline", period)
        tribal = spm_unit.household("is_on_tribal_land", period)
        p = parameters(period).gov.fcc.acp
        max_amount = (
            where(tribal, p.amount.tribal, p.amount.standard) * MONTHS_IN_YEAR
        )
        return min_(max_amount, broadband_cost)



from policyengine_us.model_api import *


class is_optional_senior_or_disabled_income_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Income-eligibility for State’s optional Medicaid pathway for seniors or people with disabilities"
    documentation = (
        "True if the tax unit’s countable income—after the state-specific "
        "income disregard—is below the income limit that the state sets for its "
        "optional pathway for aged, blind, or disabled individuals who are not "
        "otherwise SSI-eligible."
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#m"

    def formula(person, period, parameters):
        # income
        personal_income = person("ssi_countable_income", period)
        tax_unit = person.tax_unit
        income = tax_unit.sum(personal_income)
        # TODO: Each state may has its own way of calculating countable income. Not necessary follow the SSA rule.
        # Add countable income on state level, then consolidate them with this file.

        #  Flags & state info
        is_joint = tax_unit("tax_unit_is_joint", period)
        state = person.household("state_code_str", period)

        #  Parameters ─
        p = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.senior_or_disabled

        # Monthly disregard
        monthly_income_disregard = where(
            is_joint,
            p.income.disregard.couple[state],
            p.income.disregard.individual[state],
        )

        # Annualize
        income_disregard = monthly_income_disregard * MONTHS_IN_YEAR

        #  Poverty-guideline-based income limit
        limit_pct = where(
            is_joint,
            p.income.limit.couple[state],
            p.income.limit.individual[state],
        )
        fpg_annual = tax_unit("tax_unit_fpg", period)
        income_limit = limit_pct * fpg_annual

        #  Income test
        countable_income = income - income_disregard
        return countable_income < income_limit



from policyengine_us.model_api import *


class oh_additions(Variable):
    value_type = float
    entity = Person
    label = "Ohio additions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/communications/publications/individual_income_tax_ohio.pdf#page=2"
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-bundle.pdf#page=3"
        "https://cms7files1.revize.com/starkcountyoh/Document_center/Offices/Auditor/Services/Homestead%20Exemption/Ohio_Adj_Gross_Income.pdf#page=1"
    )
    defined_for = StateCode.OH
    adds = "gov.states.oh.tax.income.additions"



from policyengine_us.model_api import *


class hi_reduced_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii reduced itemized deduction"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=15"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=32"  # total itemized deduction worksheet
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    # If the state AGI of the filer exceeds a certain amount, only partial itemized deductions
    # can be deducted.
    def formula(tax_unit, period, parameters):
        total_deductions = tax_unit("hi_total_itemized_deductions", period)
        partial_deductions = add(
            tax_unit,
            period,
            [
                "hi_medical_expense_deduction",
                "investment_interest_expense",
                "hi_casualty_loss_deduction",
            ],
        )

        # eligible check 1: deduction_difference need to be greater than 0 to have reduced deduction
        partial_deductions_less_than_total = (
            partial_deductions < total_deductions
        )
        total_less_partial_ded_amount = total_deductions - partial_deductions
        # Take a percentage of the difference between the total and partial deductions
        # Hawaii applies federal limits which have been revoked in 2018
        p_2017 = parameters(
            f"2017-01-01"
        ).gov.irs.deductions.itemized.reduction
        total_less_partial_ded_percentage = (
            total_less_partial_ded_amount * p_2017.rate.base
        )
        # eligible check 2: actual AGI need to be smaller than AGI cap
        hi_agi = tax_unit("hi_agi", period)
        filing_status = tax_unit("filing_status", period)
        # Hawaii applies an federal AGI limit as of 2009.
        p_hi = parameters(
            period
        ).gov.states.hi.tax.income.deductions.itemized.threshold
        agi_threshold = p_hi.reduction[filing_status]
        agi_over_threshold = agi_threshold < hi_agi
        # If the AGI is over a threshold, the AGI amount is reduced by the threshold and multiplied
        # by a rate
        reduced_agi = hi_agi - agi_threshold
        reduced_agi_percentage = reduced_agi * p_2017.rate.excess_agi

        smaller_reduced_ded = min_(
            total_less_partial_ded_percentage, reduced_agi_percentage
        )
        reduced_deductions = max_(0, total_deductions - smaller_reduced_ded)
        eligible = partial_deductions_less_than_total & agi_over_threshold
        return eligible * reduced_deductions



from policyengine_us.model_api import *


class wv_low_income_family_tax_credit_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Adjusted gross income for the West Virginia low-income family tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-12/"  # §11-21-12 (b)
    defined_for = "wv_low_income_family_tax_credit_eligible"

    adds = [
        "adjusted_gross_income",
        "tax_exempt_interest_income",
        "wv_additions",
    ]



from policyengine_us.model_api import *


class foreign_tax_credit_potential(Variable):
    value_type = float
    entity = TaxUnit
    label = "Potential value of the Foreign tax credit"
    definition_period = YEAR
    documentation = "Foreign tax credit from Form 1116"
    unit = USD



from policyengine_us.model_api import *


class ctc_refundable_maximum(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum refundable CTC"
    unit = USD
    documentation = "The maximum refundable CTC for this person."
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
        "https://www.irs.gov/pub/irs-prior/f1040--2021.pdf",
        "https://www.irs.gov/pub/irs-prior/f1040s8--2021.pdf",
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # Use either normal or ARPA CTC maximums.
        child_amount = max_(
            person("ctc_child_individual_maximum", period),
            person("ctc_child_individual_maximum_arpa", period),
        )
        ctc = parameters(period).gov.irs.credits.ctc
        if ctc.refundable.fully_refundable:
            # Fully refundable CTC does not affect the adult CTC.
            return tax_unit.sum(child_amount)
        return tax_unit.sum(min_(child_amount, ctc.refundable.individual_max))



from policyengine_us.model_api import *


class md_tanf_maximum_benefit(Variable):
    value_type = float
    entity = SPMUnit
    label = "Maryland TANF maximum benefit"
    unit = USD
    definition_period = YEAR
    defined_for = "md_tanf_eligible"

    def formula(spm_unit, period, parameters):
        people = spm_unit("spm_unit_size", period)
        # TODO: Add DHS layer.
        p = parameters(period).gov.states.md.tanf.maximum_benefit
        capped_people = min_(people, 8).astype(int)
        additional_people = people - capped_people
        base = p.main[capped_people]
        additional_maximum_benefit = p.additional * additional_people
        monthly = base + additional_maximum_benefit
        return monthly * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class tx_regular_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = (
        "Texas Temporary Assistance for Needy Families (TANF) regular benefit"
    )
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/c-110-tanf",
    )
    defined_for = "tx_tanf_eligible"

    def formula(spm_unit, period, parameters):
        payment_standard = spm_unit("tx_tanf_payment_standard", period)
        countable_income = spm_unit("tx_tanf_countable_income", period)
        p = parameters(period).gov.states.tx.tanf

        # Calculate benefit as payment standard minus countable income
        calculated_benefit = max_(payment_standard - countable_income, 0)

        # Apply minimum grant rule
        return max_(calculated_benefit, p.minimum_grant)



from policyengine_us.model_api import *


class ut_ss_benefits_credit_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah Social Security Benefits Credit maximum amount"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = "https://incometax.utah.gov/credits/ss-benefits"

    def formula(tax_unit, period, parameters):
        """
        This credit essentially makes social security tax-exempt for people earning
        under a threshold, and taxes it back after that.
        """
        taxable_social_security = add(
            tax_unit, period, ["taxable_social_security"]
        )  # Line 1
        total_income = tax_unit("ut_total_income", period)  # Line 2
        # Skip line 3 (muncipal bond interest)
        # Skip line 4 (subtract line 3 from line 2)

        tax_exempt_interest = add(
            tax_unit, period, ["tax_exempt_interest_income"]
        )  # Line 5
        modified_agi = total_income + tax_exempt_interest  # Line 6
        p = parameters(period).gov.states.ut.tax.income
        taxed_ss = taxable_social_security * p.rate  # Line 7
        filing_status = tax_unit("filing_status", period)
        phase_out_income = max_(
            0,
            modified_agi
            - p.credits.ss_benefits.phase_out.threshold[filing_status],
        )  # Line 8
        phase_out_reduction = (
            phase_out_income * p.credits.ss_benefits.phase_out.rate
        )  # Line 9
        return max_(0, taxed_ss - phase_out_reduction)  # Line 10



from policyengine_us.model_api import *


class mt_capital_gain_credit(Variable):
    value_type = float
    entity = Person
    label = "Montana capital gain credit"
    unit = USD
    definition_period = YEAR
    reference = "https://rules.mt.gov/gateway/RuleNo.asp?RN=42%2E4%2E502"
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.credits.capital_gain

        net_capital_gain = person("capital_gains", period)
        # The net capital gain variable is capped at 0
        return p.percentage * net_capital_gain



from policyengine_us.model_api import *


class nyc_unincorporated_business_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC Unincorporated Business Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"



from policyengine_us.model_api import *


class tx_tanf_income_for_budgetary_needs_test(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas TANF income for budgetary needs test"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits",
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-408",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Budgetary needs test uses income WITHOUT earned income disregards (1/3 or 90%)
        # Deductions applied:
        # 1. Work expense ($120)
        # 2. Dependent care costs
        # 3. Child support deduction (from unearned income)

        person = spm_unit.members
        p = parameters(period).gov.states.tx.tanf.income

        # Earned income after work expense only (no 1/3 or 90% disregard)
        gross_earned = person("tx_tanf_gross_earned_income", period)
        work_expense = min_(p.deductions.work_expense, gross_earned)
        earned_after_work_expense = max_(gross_earned - work_expense, 0)

        # Sum across household members
        total_earned_after_work_expense = spm_unit.sum(
            earned_after_work_expense
        )

        # Apply dependent care deduction
        dependent_care_deduction = spm_unit(
            "tx_tanf_dependent_care_deduction", period
        )
        earned_after_deductions = max_(
            total_earned_after_work_expense - dependent_care_deduction, 0
        )

        # Add unearned income (already has child support deduction applied)
        unearned_income = spm_unit("tx_tanf_countable_unearned_income", period)

        return earned_after_deductions + unearned_income



from policyengine_us.model_api import *


class pr_refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico refundable Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.irs.gov/pub/irs-pdf/f1040s8.pdf"
        "https://www.law.cornell.edu/uscode/text/26/24#h_4_A"
    )

    # This provision is part of the federal CTC legal code
    # will will merge the logic with the federal CTC once the puerto income tax structure is completed
    def formula(tax_unit, period, parameters):
        # line 24
        # uncapped_ssi can be below 0
        ssi = max_(0, add(tax_unit, period, ["uncapped_ssi"]))
        sum_eitc_ssi = add(tax_unit, period, ["eitc"]) + ssi

        # line 25
        reduced_ss_tax = max_(
            0,
            tax_unit("pr_refundable_ctc_social_security_tax", period)
            - sum_eitc_ssi,
        )

        # line 26
        base_credit = max_(tax_unit("refundable_ctc", period), reduced_ss_tax)

        # line 27
        return min_(base_credit, tax_unit("ctc", period))



from policyengine_us.model_api import *


class ca_capi_resource_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "California CAPI resource eligible"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://www.cdss.ca.gov/Portals/9/CAPI/CAPI_Regulations-Accessible.pdf"

    def formula(spm_unit, period, parameters):
        resources = spm_unit("ca_capi_resources", period)
        p = parameters(period).gov.states.ca.cdss.capi.resources
        married = spm_unit("spm_unit_is_married", period)
        resource_limit = where(married, p.limit.couple, p.limit.single)
        return resources <= resource_limit



from policyengine_us.model_api import *


class tax_preparation_fees(Variable):
    value_type = float
    entity = Person
    label = "Tax preparation fees"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class wic(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    documentation = "Benefit value for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "WIC"
    reference = (
        "https://fns-prod.azureedge.net/sites/default/files/resource-files/WICPC2018FoodPackage-Summary.pdf#page=2",
        "https://www.law.cornell.edu/cfr/text/7/246.7",
    )
    unit = USD
    exhaustive_parameter_dependencies = "gov.usda.wic"
    defined_for = "is_wic_eligible"

    def formula(person, period, parameters):
        category = person("wic_category_str", period)
        p = parameters(period).gov.usda.wic
        values = p.value
        value_if_eligible = values[category]
        would_takeup = person("would_claim_wic", period)
        if p.abolish_wic:
            return 0
        return would_takeup * value_if_eligible



from policyengine_us.model_api import *


class medicaid_magi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Medicaid/CHIP/ACA-related Modified AGI"
    unit = USD
    documentation = "Medicaid/CHIP/ACA-related modified AGI for this tax unit."
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/42/1396a#e_14_G",  # Medicaid law pointing to IRC
        "https://www.law.cornell.edu/uscode/text/26/36B#d_2",  # IRC defining income
    )

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        agi_additions = parameters(period).gov.hhs.medicaid.income.modification
        return max_(0, agi + add(tax_unit, period, agi_additions))



from policyengine_us.model_api import *


class oh_insured_unreimbursed_medical_care_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "Ohio insured unreimbursed medical and health care expense deduction"
    )
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=18",  # Line 36
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=27",
        "https://codes.ohio.gov/ohio-revised-code/section-5747.01",  # R.C. 5747.01(10)
    )
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        # Line 5
        total_expenses = add(
            tax_unit,
            period,
            ["oh_insured_unreimbursed_medical_care_expense_amount"],
        )
        # Line 6
        federal_agi = tax_unit("adjusted_gross_income", period)

        # Can deduct medical expenses in excess of 7.5% of federal AGI.
        # Line 7
        rate = parameters(
            period
        ).gov.states.oh.tax.income.deductions.unreimbursed_medical_care_expenses.rate
        agi_floor = federal_agi * rate
        # Line 8
        return max_(0, total_expenses - agi_floor)



from policyengine_us.model_api import *


class de_elderly_or_disabled_income_exclusion_joint(Variable):
    value_type = float
    entity = Person
    label = "Delaware individual aged or disabled exclusion when married filing jointly"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf#page=1"
    defined_for = "de_elderly_or_disabled_income_exclusion_eligible_person"

    def formula(person, period, parameters):
        tax_unit = person.tax_unit
        # First get their filing status.
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.de.tax.income.subtractions.exclusions.elderly_or_disabled

        # If filing jointly, both spouses have to be eligible.
        joint = filing_status == filing_status.possible_values.JOINT

        head = person("is_tax_unit_head", period)

        spouse = person("is_tax_unit_spouse", period)

        eligible_person = person(
            "de_elderly_or_disabled_income_exclusion_eligible_person", period
        )

        eligible_head_present = tax_unit.any(head & eligible_person)

        eligible_spouse_present = tax_unit.any(spouse & eligible_person)

        eligible_joint_unit = eligible_head_present & eligible_spouse_present

        eligible_unit = where(joint, eligible_joint_unit, eligible_person)
        return p.amount[filing_status] * eligible_unit



from policyengine_us.model_api import *


class specified_possession_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income from Guam, American Samoa, or the Northern Mariana Islands"
    unit = USD
    documentation = "Income generated in the above territories by any individual who is a bona fide resident."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/931"



from policyengine_us.model_api import *


class ar_exemptions(Variable):
    value_type = float
    entity = Person
    label = "Arkansas exemptions from income tax for each individual"
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=10"
    defined_for = StateCode.AR
    unit = USD
    definition_period = YEAR
    adds = "gov.states.ar.tax.income.exemptions.exemptions"



from policyengine_us.model_api import *


class nm_deduction_for_certain_dependents(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico deduction for certain dependents"
    unit = USD
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503892/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgAcATgBMASgA0ybKUIQAiokK4AntADkW6REJhcCFWs069BoyADKeUgCFNAJQCiAGRcA1AIIA5AMIu0qRgAEbQpOySkkA"
    defined_for = "nm_deduction_for_certain_dependents_eligible"

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.nm.tax.income.deductions.certain_dependents
        # The law 7-2-39(D) defines dependents as those from IRC 152.
        # IRC 152 refers to all dependents.
        # https://www.law.cornell.edu/uscode/text/26/152
        dependents = tax_unit("tax_unit_dependents", period)
        # New Mexico reduces the number of claimable dependents by one.
        countable_dependents = max_(dependents - 1, 0)
        amount_per_dependent = p.amount[filing_status]
        return amount_per_dependent * countable_dependents



from policyengine_us.model_api import *


class ks_total_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas total EITC amount (both nonrefundable and refundable)"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.ks.tax.income.credits
        return p.eitc_fraction * federal_eitc



from policyengine_us.model_api import *


class vt_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont child tax credit"
    definition_period = YEAR
    unit = USD
    reference = "https://casetext.com/statute/vermont-statutes/title-32-taxation-and-finance/chapter-151-income-taxes/subchapter-002-taxation-of-individuals-trusts-and-estates/section-5830f-see-note-vermont-child-tax-credit/1"
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        # Get age status of all people in the tax unit.
        person = tax_unit.members
        age = person("age", period)
        p = parameters(period).gov.states.vt.tax.income.credits.ctc
        eligible = age <= p.age_limit
        count_eligible = tax_unit.sum(eligible)
        # Get maximum credit amount.
        max_credit = p.amount * count_eligible
        # Get adjusted gross income.
        agi = tax_unit("adjusted_gross_income", period)
        # Reduce credit amount over the phaseout range.
        excess_agi = max_(agi - p.reduction.start, 0)
        increments = np.ceil(excess_agi / p.reduction.increment)
        total_reduction = p.reduction.amount * increments
        # Return reduced credit amount.
        return max_(max_credit - total_reduction, 0)



from policyengine_us.model_api import *


class az_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Property Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "az_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(period).gov.states.az.tax.income.credits.property_tax
        income = tax_unit("az_property_tax_credit_income", period)

        cohabitating = tax_unit("cohabitating_spouses", period)

        cap = where(
            cohabitating,
            p.amount.cohabitating.calc(income),
            p.amount.living_alone.calc(income),
        )

        # Rent counts equally to property taxes.
        property_tax_plus_rent = add(
            tax_unit, period, ["real_estate_taxes", "rent"]
        )
        # property_tax and rent are equally weighted

        return min_(cap, property_tax_plus_rent)



from policyengine_us.model_api import *


class ct_income_tax_phase_out_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut income tax phase out add back"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        income = tax_unit("ct_agi", period)
        p = parameters(period).gov.states.ct.tax.income.add_back
        filing_status = tax_unit("filing_status", period)
        start = p.start[filing_status]
        max_amount = p.max_amount[filing_status]
        increment = p.increment[filing_status]
        reduction_per_increment = p.amount[filing_status]
        excess = max_(income - start, 0)
        increments = np.ceil(excess / increment)
        amount = increments * reduction_per_increment
        return min_(max_amount, amount)



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_senior_deduction_extension() -> Reform:
    def modify_parameters(parameters):
        # Update deductions_if_itemizing to include additional_senior_deduction after 2028
        parameters.gov.irs.deductions.deductions_if_itemizing.update(
            start=instant("2029-01-01"),
            stop=instant("2099-12-31"),
            value=[
                "qualified_business_income_deduction",
                "wagering_losses_deduction",
                "itemized_taxable_income_deductions",
                "additional_senior_deduction",
            ],
        )

        # Update deductions_if_not_itemizing to include additional_senior_deduction after 2028
        parameters.gov.irs.deductions.deductions_if_not_itemizing.update(
            start=instant("2029-01-01"),
            stop=instant("2099-12-31"),
            value=[
                "standard_deduction",
                "qualified_business_income_deduction",
                "additional_senior_deduction",
                "charitable_deduction_for_non_itemizers",
            ],
        )

        return parameters

    class reform(Reform):
        def apply(self):
            self.modify_parameters(modify_parameters)

    return reform


def create_senior_deduction_extension_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_senior_deduction_extension()

    p = parameters.gov.contrib.crfb.senior_deduction_extension

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).applies:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_senior_deduction_extension()
    else:
        return None


senior_deduction_extension = create_senior_deduction_extension_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class oh_modified_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio modified adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=31",
    )
    defined_for = StateCode.OH

    adds = ["oh_agi"]


# Additionally adds the Ohio Business income deduction



from policyengine_us.model_api import *


class tax_unit_unemployment_compensation(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit unemployment compensation"
    unit = USD
    documentation = "Combined unemployment compensation for the tax unit."
    definition_period = YEAR

    adds = ["unemployment_compensation"]



from policyengine_us.model_api import *


class qbid_amount(Variable):
    value_type = float
    entity = Person
    label = (
        "Per-cap qualified business income deduction amount for each person"
    )
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/199A#b_1"
        "https://www.irs.gov/pub/irs-prior/p535--2018.pdf"
    )

    def formula(person, period, parameters):
        # computations follow logic in 2018 IRS Publication 535,
        # Worksheet 12-A (and Schedule A for SSTB)
        p = parameters(period).gov.irs.deductions.qbi
        # compute maximum QBID amount
        qbi = person("qualified_business_income", period)
        qbid_max = p.max.rate * qbi  # Worksheet 12-A, line 3
        # compute caps
        w2_wages = person("w2_wages_from_qualified_business", period)
        b_property = person("unadjusted_basis_qualified_property", period)
        wage_cap = w2_wages * p.max.w2_wages.rate  # Worksheet 12-A, line 5
        alt_cap = (  # Worksheet 12-A, line 9
            w2_wages * p.max.w2_wages.alt_rate
            + b_property * p.max.business_property.rate
        )
        full_cap = max_(wage_cap, alt_cap)  # Worksheet 12-A, line 10
        # compute phase-out ranges
        taxinc_less_qbid = person.tax_unit("taxable_income_less_qbid", period)
        filing_status = person.tax_unit("filing_status", period)
        po_start = p.phase_out.start[filing_status]
        po_length = p.phase_out.length[filing_status]
        # compute phase-out limited QBID amount
        reduction_rate = min_(  # Worksheet 12-A, line 24; Schedule A, line 9
            1, (max_(0, taxinc_less_qbid - po_start)) / po_length
        )
        applicable_rate = 1 - reduction_rate  # Schedule A, line 10
        is_sstb = person("business_is_sstb", period)
        # Schedule A, line 11
        sstb_multiplier = where(is_sstb, applicable_rate, 1)
        adj_qbid_max = qbid_max * sstb_multiplier
        # Schedule A, line 12 and line 13
        adj_cap = full_cap * sstb_multiplier
        line11 = min_(adj_qbid_max, adj_cap)  # Worksheet 12-A, line 11
        # compute phased reduction
        reduction = reduction_rate * max_(  # Worksheet 12-A, line 25
            0, adj_qbid_max - adj_cap
        )
        line26 = max_(0, adj_qbid_max - reduction)
        line12 = where(adj_cap < adj_qbid_max, line26, 0)
        return max_(line11, line12)  # Worksheet 12-A, line 13



from policyengine_us.model_api import *


class dc_ccsp_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible child for DC Child Care Subsidy Program (CCSP)"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=8"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.age_threshold
        age = person("monthly_age", period)
        is_disabled = person("is_disabled", period)
        age_limit = where(is_disabled, p.disabled_child, p.child)
        age_eligible = age < age_limit
        is_dependent = person("is_tax_unit_dependent", period)
        immigration_status_eligible = person(
            "dc_ccsp_immigration_status_eligible_person", period
        )

        return age_eligible & is_dependent & immigration_status_eligible



from policyengine_us.model_api import *


class mt_disability_income_exclusion_person(Variable):
    value_type = float
    entity = Person
    label = "Montana disability income exclusion for each person"
    defined_for = "mt_disability_income_exclusion_eligible_person"
    unit = USD
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf#page=31"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.subtractions.disability_income
        return min_(p.cap, person("disability_benefits", period))



from policyengine_us.model_api import *


class la_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana adjusted gross income income"
    defined_for = StateCode.LA
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        exempt_income = tax_unit("la_agi_exempt_income", period)
        return max_(agi - exempt_income, 0)



from policyengine_us.model_api import *


class tx_lifeline_supplement(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas Lifeline supplement amount"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.TX
    reference = "https://www.law.cornell.edu/regulations/texas/16-Tex-Admin-Code-SS-26-412"

    adds = ["gov.states.tx.uct.lifeline.supplement"]



from policyengine_us.model_api import *


class ma_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    reference = "https://malegislature.gov/Laws/GeneralLaws/PartI/TitleIX/Chapter62/Section4"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        tax_before_credits = tax_unit("ma_income_tax_before_credits", period)
        non_refundable_credits = tax_unit("ma_non_refundable_credits", period)
        return max_(tax_before_credits - non_refundable_credits, 0)



from policyengine_us.model_api import *


class state_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "State adjusted gross income"
    unit = USD
    definition_period = YEAR
    adds = "gov.states.household.state_agis"

    def formula(tax_unit, period, parameters):
        # States that adopt the federal AGI
        # Based on comments in state_agis.yaml
        FEDERAL_AGI_STATES = [
            "CO",  # Colorado
            "MI",  # Michigan
            "MN",  # Minnesota
            "NC",  # North Carolina
            "ND",  # North Dakota
            "NM",  # New Mexico
            "SC",  # South Carolina
        ]

        # Get the current state
        state_code = tax_unit.household("state_code_str", period)

        # Get the sum of state-specific AGIs
        state_specific_base = add(
            tax_unit,
            period,
            parameters(period).gov.states.household.state_agis,
        )

        # Special handling for states that have separate individual and joint AGI calculations
        # Arkansas and Delaware use different AGI calculations based on filing status
        STATES_WITH_INDIVIDUAL_JOINT_AGI = ["AR", "DE"]

        # Calculate maximum AGI for states with individual/joint distinction
        # This ensures we use the higher of the two calculations
        for state in STATES_WITH_INDIVIDUAL_JOINT_AGI:
            is_state = state_code == state
            indiv_var = f"{state.lower()}_agi_indiv"
            joint_var = f"{state.lower()}_agi_joint"
            indiv_agi = add(tax_unit, period, [indiv_var])
            joint_agi = add(tax_unit, period, [joint_var])
            max_agi = max_(indiv_agi, joint_agi)
            state_specific_base = where(is_state, max_agi, state_specific_base)

        # Check if the state adopts federal AGI
        uses_federal = np.isin(state_code, FEDERAL_AGI_STATES)

        # Get federal AGI
        federal_agi = tax_unit("adjusted_gross_income", period)

        # Return federal AGI for states that adopt it, otherwise state-specific
        return where(uses_federal, federal_agi, state_specific_base)



from policyengine_us.model_api import *


class ma_eaedc_countable_assets(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC countable assets"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.MA

    adds = "gov.states.ma.dta.tcap.eaedc.assets.sources"



from policyengine_us.model_api import *


class tx_dart_reduced_fare_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Eligible for Dallas Area Rapid Transit (DART) Reduced Fare program"
    )
    definition_period = YEAR
    defined_for = StateCode.TX
    reference = (
        "https://www.dart.org/fare/general-fares-and-overview/reduced-fares"
    )

    def formula(person, period, parameters):
        # Eligible due to age
        age_eligible = person("tx_dart_reduced_fare_age_eligible", period)
        # Disability
        is_disabled = person("is_disabled", period)
        # Veteran
        veteran_eligible = person("is_veteran", period)
        # Student (high school, college, or trade school)
        student_eligible = person("is_full_time_student", period)
        # Enrolled in applicable programs (Discount GoPass)
        # Note: DART has two separate programs that both offer 50% discounts:
        # 1. Reduced Fare program (for age/disability/veteran/student)
        # 2. Discount GoPass program (for assistance program recipients)
        # We combine them here since they provide identical benefits
        enrolled_eligible = person(
            "tx_dart_reduced_fare_program_eligible", period
        )

        return (
            age_eligible
            | is_disabled
            | veteran_eligible
            | student_eligible
            | enrolled_eligible
        )



from policyengine_us.model_api import *


class il_liheap_crisis_assistance_amount(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois LIHEAP crisis assistance amount"
    unit = USD
    definition_period = YEAR
    defined_for = "il_liheap_eligible_for_crisis_assistance"
    reference = "https://dceo.illinois.gov/communityservices/utilitybillassistance.html"

    adds = ["gov.states.il.dceo.liheap.payment.crisis_amount.max"]



from policyengine_us.model_api import *


class in_nol(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana NOL"
    unit = USD
    definition_period = YEAR
    documentation = "Net operating losses allowable for deduction in Indiana."
    reference = (
        "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-2-2.5"
    )
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class la_disability_income_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "Louisiana disability income exemption for each person"
    unit = USD
    definition_period = YEAR
    reference = "https://www.legis.la.gov/legis/Law.aspx?d=102133"  # (B)
    defined_for = StateCode.LA

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.la.tax.income.exempt_income.disability
        # The disability_benefits variable includes only taxable disability benefits.
        disability_benefits = person("disability_benefits", period)
        capped_disability_benefits = min_(disability_benefits, p.cap)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        total_disablity_exemption = (
            capped_disability_benefits * is_head_or_spouse
        )
        # People who reveive the blind exemption are not eligible for
        # the disability income exemption
        blind_exemption_received = person(
            "la_receives_blind_exemption", period
        )
        return where(blind_exemption_received, 0, total_disablity_exemption)



from policyengine_us.model_api import *


class az_blind_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona blind exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.az.tax.income.exemptions

        blind_head = tax_unit("blind_head", period)
        blind_spouse = tax_unit("blind_spouse", period)

        count_eligible = blind_head.astype(int) + blind_spouse.astype(int)
        return count_eligible * p.blind



from policyengine_us.model_api import *


class tax_unit_social_security(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit Social Security"
    unit = USD
    definition_period = YEAR

    adds = ["social_security"]



from policyengine_us.model_api import *


class mt_applicable_ald_deductions(Variable):
    value_type = float
    entity = Person
    label = "Montana applicable above-the-line deductions "
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        ald_deductions = person.tax_unit("above_the_line_deductions", period)
        total_deductions = ald_deductions / person.tax_unit(
            "head_spouse_count", period
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return head_or_spouse * total_deductions



from policyengine_us.model_api import *


class OregonWFHDCEligibilityCategory(Enum):
    YOUNGEST = "Youngest"  # under 3 years old
    YOUNG = "Young"  # between 3 and 6 years old
    OLD = "Old"  # between 6 and 13 years old
    DISABLED_TEENS = "Disabled teenagers"  # disabled teenagers that are between 13 and 18 years old
    DISABLED_ADULTS = (
        "Disabled adults"  # disabled adults that are over 18 years old
    )
    NONE = "None"


class or_wfhdc_eligibility_category(Variable):
    value_type = Enum
    possible_values = OregonWFHDCEligibilityCategory
    entity = TaxUnit
    label = "Oregon working family household and dependent care credit percentage table column"
    definition_period = YEAR
    defined_for = "or_wfhdc_eligible"
    reference = "https://www.oregon.gov/dor/forms/FormsPubs/publication-or-wfhdc-tb_101-458_2021.pdf#page=1"
    default_value = OregonWFHDCEligibilityCategory.NONE

    def formula(tax_unit, period, parameters):
        # Column determined by age of youngest child and whether they have a disability.

        p = (
            parameters(period)
            .gov.states["or"]
            .tax.income.credits.wfhdc.age_range
        )
        # Get the age of the youngest qualifying child.
        person = tax_unit.members
        age = person("age", period)
        min_age = tax_unit.min(age)

        # Determine if the youngest qualifying individual is disabled.
        # The household has at least one qualifying individual because they are WFHDC eligible.
        disabled = person("is_disabled", period)
        disabled_rank = person.get_rank(tax_unit, age, disabled)
        youngest_and_disabled = disabled_rank == -1

        # This will be true if any child with the lowest age is disabled.
        youngest_is_disabled = tax_unit.sum(youngest_and_disabled) > 0
        conditions = [
            min_age < p.youngest,
            min_age < p.young,
            min_age < p.old,
            (min_age < p.oldest) & youngest_is_disabled,
            youngest_is_disabled,
        ]
        values = [
            OregonWFHDCEligibilityCategory.YOUNGEST,
            OregonWFHDCEligibilityCategory.YOUNG,
            OregonWFHDCEligibilityCategory.OLD,
            OregonWFHDCEligibilityCategory.DISABLED_TEENS,
            OregonWFHDCEligibilityCategory.DISABLED_ADULTS,
        ]
        return select(
            conditions, values, default=OregonWFHDCEligibilityCategory.NONE
        )



from policyengine_us.model_api import *


class taxsim_year(Variable):
    value_type = int
    entity = TaxUnit
    label = "Policy year"
    documentation = "Tax year ending Dec 31(4 digits between 1960 and 2023, but state must be zero if year is before 1977. (We don't have code for state laws before 1977.) State tax laws are effectively inflated by 2.5%/year after 2021."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return period.start.year



from policyengine_us.model_api import *


class ma_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = "https://malegislature.gov/Laws/GeneralLaws/PartI/TitleIX/Chapter62/Section4"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        part_a_dividends = tax_unit(
            "ma_part_a_taxable_dividend_income", period
        )
        part_a_capital_gains = tax_unit(
            "ma_part_a_taxable_capital_gains_income", period
        )
        part_b = tax_unit("ma_part_b_taxable_income", period)
        part_c = tax_unit("ma_part_c_taxable_income", period)
        rates = parameters(period).gov.states.ma.tax.income.rates
        exempt = tax_unit("is_ma_income_tax_exempt", period)
        tax_on_income = (
            rates.part_a.dividends * part_a_dividends
            + rates.part_a.capital_gains * part_a_capital_gains
            + rates.part_b * part_b
            + rates.part_c * part_c
        )
        total_taxable_income = (
            part_a_dividends + part_a_capital_gains + part_b + part_c
        )
        additional_tax = rates.additional.calc(total_taxable_income)
        return where(exempt, 0, tax_on_income + additional_tax)



from policyengine_us.model_api import *


class dc_ccsp_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for DC Child Care Subsidy Program (CCSP) based on immigration status"
    definition_period = MONTH
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=8"
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        return np.isin(
            immigration_status_str,
            p.qualified_immigration_statuses,
        )



from policyengine_us.model_api import *


class educator_expense(Variable):
    value_type = float
    entity = Person
    label = "Educator expenses"
    unit = USD
    documentation = (
        "Expenses necessary for carrying out educator-related duties."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class rrc_arpa(Variable):
    value_type = float
    entity = TaxUnit
    label = "Recovery Rebate Credit (ARPA)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/6428B"

    def formula(tax_unit, period, parameters):
        rrc = parameters(period).gov.irs.credits.recovery_rebate_credit
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        count_dependents = tax_unit("tax_unit_count_dependents", period)
        count_adults = where(tax_unit("tax_unit_is_joint", period), 2, 1)
        max_payment = (
            rrc.arpa.max.adult * count_adults
            + rrc.arpa.max.dependent * count_dependents
        )
        phase_out_length = rrc.arpa.phase_out.length[filing_status]
        excess = max_(0, agi - rrc.arpa.phase_out.threshold[filing_status])
        payment_reduction_percent = excess / phase_out_length
        payment_reduction = max_payment * payment_reduction_percent
        return max_(0, max_payment - payment_reduction)



from policyengine_us.model_api import *


class tuition_and_fees_deduction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Tuition and fees deduction eligible"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.irs.gov/pub/irs-pdf/f8917.pdf#page=2"
        # Law was repealed in 2021
        "https://irc.bloombergtax.com/public/uscode/doc/irc/section_222"
    )

    def formula(tax_unit, period, parameters):
        # Married filing separately are not eligible for this deduction.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # Can't claim this deduction if the household has taken the
        # American Oppportunity or Lifetime Learning Credit.
        aoc_llc = add(
            tax_unit,
            period,
            [
                "american_opportunity_credit",
                "lifetime_learning_credit_potential",
            ],
        )
        return ~separate & (aoc_llc == 0)



from policyengine_us.model_api import *


class snap_excluded_child_earner(Variable):
    value_type = bool
    entity = Person
    label = "Excluded child earner"
    documentation = "Whether this person is a child whose earned income is excluded from SNAP "
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#d_7"

    def formula(person, period, parameters):
        p = p = parameters(period).gov.usda.snap.income
        age = person("monthly_age", period)
        is_in_k12_school = person("is_in_k12_school", period)
        age_eligible = age <= p.child_income_exclusion_age
        return is_in_k12_school & age_eligible



from policyengine_us.model_api import *


class la_dependents_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana qualified dependents exemption"
    reference = [
        "https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf#page=1",
        "https://revenue.louisiana.gov/TaxForms/IT540iWEB(2022)D1.pdf#page=2",
    ]
    # Even though the tax computation worksheet refers "dependent exemption" as credits, the instructions for
    # preparing tax form line 6a-6b specifies it as exemption.
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        dependents = tax_unit("tax_unit_dependents", period)
        p = parameters(period).gov.states.la.tax.income.exemptions
        return dependents * p.dependent



from policyengine_us.model_api import *


class ma_tafdc_immigration_status_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) due to immigration status"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-400"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        immigration_status = spm_unit.members("immigration_status", period)
        undocumented = (
            immigration_status
            == immigration_status.possible_values.UNDOCUMENTED
        )
        return spm_unit.any(~undocumented)



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def non_refundable_ss_credit_reform() -> Reform:
    class ss_credit(Variable):
        value_type = float
        entity = Person
        label = "Social Security credit"
        definition_period = YEAR
        unit = USD

        def formula(person, period, parameters):
            taxable_social_security = person("taxable_social_security", period)
            highest_tax_rate = person.tax_unit("highest_tax_rate", period)
            ss_tax = taxable_social_security * highest_tax_rate
            p = parameters(period).gov.contrib.crfb.ss_credit
            filing_status = person.tax_unit("filing_status", period)
            return min_(ss_tax, p.amount[filing_status])

    class highest_tax_rate(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Highest tax rate faced by tax unit"
        reference = "https://www.law.cornell.edu/uscode/text/26/1"
        unit = "/1"

        def formula(tax_unit, period, parameters):
            # compute taxable income that is taxed at the main rates
            full_taxable_income = tax_unit("taxable_income", period)
            cg_exclusion = tax_unit(
                "capital_gains_excluded_from_taxable_income", period
            )
            taxinc = max_(0, full_taxable_income - cg_exclusion)

            # get bracket rates and thresholds
            p = parameters(period).gov.irs.income
            filing_status = tax_unit("filing_status", period)

            # Vectorized approach: determine the highest applicable rate for each household
            # We'll check each bracket and keep the rate if income falls within it

            # Start with the base rate (10%)
            highest_rate = p.bracket.rates["1"]

            # Get all bracket numbers
            bracket_numbers = list(p.bracket.rates.__iter__())

            # Process each bracket from lowest to highest
            # The threshold is the upper bound of each bracket
            prev_threshold = np.zeros_like(taxinc)

            for bracket_num in bracket_numbers:
                b = str(bracket_num)
                rate = p.bracket.rates[b]

                # Get the upper threshold for this bracket based on filing status
                if b == "7":
                    # The top bracket has no upper limit
                    threshold = np.inf * np.ones_like(taxinc)
                else:
                    threshold = p.bracket.thresholds[b][filing_status]

                # Update the rate for households with income in this bracket
                # Income is in this bracket if it's > prev_threshold and <= threshold
                in_bracket = (taxinc > prev_threshold) & (taxinc <= threshold)
                highest_rate = where(in_bracket, rate, highest_rate)

                # Store this threshold as the previous for next iteration
                prev_threshold = where(
                    threshold != np.inf, threshold, prev_threshold
                )

            return highest_rate

    def modify_parameters(parameters):
        parameters.gov.irs.credits.non_refundable.update(
            start=instant("2026-01-01"),
            stop=instant("2035-12-31"),
            value=[
                "foreign_tax_credit",
                "cdcc",
                "non_refundable_american_opportunity_credit",
                "lifetime_learning_credit",
                "savers_credit",
                "residential_clean_energy_credit",
                "energy_efficient_home_improvement_credit",
                "elderly_disabled_credit",
                "new_clean_vehicle_credit",
                "used_clean_vehicle_credit",
                "non_refundable_ctc",
                "ss_credit",
            ],
        )
        return parameters

    class reform(Reform):
        def apply(self):
            self.update_variable(ss_credit)
            self.update_variable(highest_tax_rate)
            self.modify_parameters(modify_parameters)

    return reform


def create_non_refundable_ss_credit_reform(
    parameters, period, bypass: bool = False
):
    # Create a create_{reform name} function that initializes the reform object
    # There are two sufficient conditions for this function to return
    # the reform

    # 1. If bypass is set to true
    if bypass is True:
        return non_refundable_ss_credit_reform()

    # 2. If boolean in in_effect.yaml is set to true
    parameter = parameters.gov.contrib.crfb.ss_credit
    current_period = period_(period)
    reform_active = False

    for i in range(5):
        if parameter(current_period).in_effect:
            # If in any of the next five years, the boolean is true,
            # set the boolean reform_active to true, and stop the check,
            # i.e., assume the reform is active in all subsequent years.
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    # if the loop set reform_active to true, return the reform.
    if reform_active:
        return non_refundable_ss_credit_reform()
    else:
        return None


# Create a reform object to by setting bypass to true,
# for the purpose of running tests
non_refundable_ss_credit_reform_object = (
    create_non_refundable_ss_credit_reform(None, None, bypass=True)
)



from policyengine_us.model_api import *


class tx_fpp_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Texas Family Planning Program eligibility"
    definition_period = YEAR
    reference = (
        "https://www.healthytexaswomen.org/healthcare-programs/family-planning-program/fpp-who-can-apply",
        "https://www.hhs.texas.gov/sites/default/files/documents/texas-womens-health-programs-report-2024.pdf",
    )
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        age_eligible = person("tx_fpp_age_eligible", period)
        income_eligible = person.spm_unit("tx_fpp_income_eligible", period)
        return age_eligible & income_eligible



from policyengine_us.model_api import *


class residential_efficiency_electrification_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Residential efficiency and electrification rebate"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.doe.residential_efficiency_electrification_rebate
        expenditures = tax_unit(
            "residential_efficiency_electrification_retrofit_expenditures",
            period,
        )
        savings_kwh = tax_unit(
            "residential_efficiency_electrification_retrofit_energy_savings",
            period,
        )
        current_kwh = tax_unit.household("current_home_energy_use", period)
        # avoid array divide-by-zero warnings by not using where() function
        savings_pct = np.zeros_like(current_kwh)
        mask = current_kwh > 0
        savings_pct[mask] = savings_kwh[mask] / current_kwh[mask]
        income_ami = tax_unit.household("household_income_ami_ratio", period)
        high_cap = p.cap.high.calc(income_ami)
        medium_cap = p.cap.medium.calc(income_ami)
        # Low cap is a dollar amount per given percentage reduction of energy
        # use per dwelling unit for the average home in the state
        average_home_energy_use_in_state = tax_unit.household(
            "average_home_energy_use_in_state", period
        )
        low_cap_per_percent = p.cap.low.calc(income_ami)
        # avoid array divide-by-zero warnings by not using where() function
        reduction = np.zeros_like(average_home_energy_use_in_state)
        mask = average_home_energy_use_in_state > 0
        reduction[mask] = (
            low_cap_per_percent[mask] / average_home_energy_use_in_state[mask]
        )
        low_cap_per_kwh_reduction = 100 * reduction
        low_cap = low_cap_per_kwh_reduction * savings_kwh
        # Uncapped amount is a percent of project costs
        percent = p.percent.calc(income_ami)
        uncapped = percent * expenditures
        cap = select(
            [
                savings_pct >= p.threshold.high,
                savings_pct >= p.threshold.medium,
                savings_pct >= p.threshold.low,
            ],
            [high_cap, medium_cap, low_cap],
            default=0,
        )
        return min_(uncapped, cap)



from policyengine_us.model_api import *


class snap_excess_medical_expense_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP excess medical expense deduction"
    unit = USD
    documentation = (
        "Deduction from SNAP gross income for excess medical expenses"
    )
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_5"

    def formula(spm_unit, period, parameters):
        # Deduction applies to medical expenses incurred by elderly or disabled
        # members only.
        person = spm_unit.members
        elderly = person("is_usda_elderly", period)
        disabled = person("is_usda_disabled", period)
        moop = person("medical_out_of_pocket_expenses", period)
        elderly_disabled_moop = spm_unit.sum(moop * (elderly | disabled))
        p = parameters(
            period
        ).gov.usda.snap.income.deductions.excess_medical_expense
        excess = max_(elderly_disabled_moop - p.disregard, 0)
        # Calculate standard medical deduction (SMD).
        state = spm_unit.household("state_code_str", period)
        standard = p.standard[state]
        standard_claimable = where(excess > 0, standard, 0)
        # Return the greater of SMD and normal deduction.
        return max_(excess, standard_claimable)



from policyengine_us.model_api import *


class ct_income_tax_after_personal_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut income tax after personal tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("ct_taxable_income", period)
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        p = parameters(period).gov.states.ct.tax.income.rates
        itax_before_personal_credits = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.SEPARATE,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.single.calc(taxable_income),
                p.joint.calc(taxable_income),
                p.separate.calc(taxable_income),
                p.surviving_spouse.calc(taxable_income),
                p.head_of_household.calc(taxable_income),
            ],
        )
        add_back = tax_unit("ct_income_tax_phase_out_add_back", period)
        tax_recapture = tax_unit("ct_income_tax_recapture", period)
        total = itax_before_personal_credits + add_back + tax_recapture
        personal_credit_rate = tax_unit("ct_personal_credit_rate", period)
        personal_credit_amount = personal_credit_rate * total
        return max_(0, total - personal_credit_amount)



from policyengine_us.model_api import *


class ma_tafdc_applicable_income_for_financial_eligibility(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "Applicable income for the Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) financial eligibility check"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-280"
    )
    defined_for = StateCode.MA

    adds = [
        "ma_tafdc_earned_income_after_deductions",
        "ma_tafdc_countable_unearned_income",
    ]



from policyengine_us.model_api import *


class mt_capital_gains_tax_joint(Variable):
    value_type = float
    entity = Person
    label = "Montana net long-term capital gains tax when married couples file jointly"
    unit = USD
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/12/Form_2_2023_Instructions.pdf#page=6"  # Net Long-Term Capital Gains Tax Table
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.main.capital_gains
        # the tax for capital gains comes into effect after 2024
        if p.in_effect:
            # Line instructions from the 2024 Montana Individual Income Tax Return Form 2
            # https://revenue.mt.gov/files/Forms/Montana-Individual-Income-Tax-Return-Form-2/2024_Montana_Individual_Income_Tax_Return_Form_2.pdf#page=2
            # Line 1
            taxable_income = person("mt_taxable_income_joint", period)
            # Line 2
            capital_gains = person("long_term_capital_gains", period)
            # Line 3
            lesser_of_cg_and_taxable_income = min_(
                capital_gains, taxable_income
            )
            # Line 4
            excess_over_taxable_income = max_(
                taxable_income - lesser_of_cg_and_taxable_income, 0
            )
            filing_status = person.tax_unit("filing_status", period)
            # Line 5
            applicable_threshold = person(
                "mt_capital_gains_tax_applicable_threshold_joint", period
            )
            status = filing_status.possible_values

            lower_rate = select(
                [
                    filing_status == status.SINGLE,
                    filing_status == status.SEPARATE,
                    filing_status == status.JOINT,
                    filing_status == status.SURVIVING_SPOUSE,
                    filing_status == status.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.rates.single.amounts[0],
                    p.rates.separate.amounts[0],
                    p.rates.joint.amounts[0],
                    p.rates.surviving_spouse.amounts[0],
                    p.rates.head_of_household.amounts[0],
                ],
            )
            higher_rate = select(
                [
                    filing_status == status.SINGLE,
                    filing_status == status.SEPARATE,
                    filing_status == status.JOINT,
                    filing_status == status.SURVIVING_SPOUSE,
                    filing_status == status.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.rates.single.amounts[-1],
                    p.rates.separate.amounts[-1],
                    p.rates.joint.amounts[-1],
                    p.rates.surviving_spouse.amounts[-1],
                    p.rates.head_of_household.amounts[-1],
                ],
            )
            # Line 6
            excess_over_threshold = max_(
                applicable_threshold - excess_over_taxable_income, 0
            )
            # Line 7
            capital_gains_below_threshold = min_(
                excess_over_threshold, lesser_of_cg_and_taxable_income
            )
            # Line 8
            lower_capital_gains_tax = (
                capital_gains_below_threshold * lower_rate
            )
            # Line 9
            income_above_threshold = max_(
                lesser_of_cg_and_taxable_income - excess_over_threshold, 0
            )
            # Line 10
            higher_capital_gains_tax = income_above_threshold * higher_rate
            # Line 11
            return lower_capital_gains_tax + higher_capital_gains_tax

        return 0



from policyengine_us.model_api import *


class heat_pump_water_heater_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Expenditures on heat pump water heaters"
    unit = USD
    definition_period = YEAR
    reference = (
        # Pre-IRA: Electric only.
        "https://www.law.cornell.edu/uscode/text/26/25C#d_3",
        # Post-IRA: Electric or natural gas.
        "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=339#342",
    )



from policyengine_us.model_api import *


class basic_income_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic income phase-out"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        basic_income = add(tax_unit, period, ["basic_income_before_phase_out"])
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        if bi.taxable:
            return 0
        p = bi.phase_out
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        threshold = p.threshold[filing_status]
        income_over_threshold = max_(0, agi - threshold)
        # Phase out as a rate above the threshold if selected.
        if p.by_rate:
            return min_(income_over_threshold * p.rate, basic_income)
        # Otherwise phase out linearly until the end of the region.
        phase_out_width = p.end[filing_status] - threshold
        pct_phase_out = min_(income_over_threshold / phase_out_width, 1)
        return basic_income * pct_phase_out



from policyengine_us.model_api import *


class me_itemized_deductions_pre_phaseout(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine itemized deductions before phaseout"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_2_ff.pdf",
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5125.html",
    )
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):

        # Get the Maine itemizing deduction parameters.
        p = parameters(period).gov.states.me.tax.income.deductions.itemized

        # Get federal itemized deductions minus SALT (Section 3A).
        deduction = tax_unit("itemized_deductions_less_salt", period)

        # Get medical (and dental) expenses.
        medical_expenses = tax_unit("medical_expense_deduction", period)

        # Get real estate (and property) taxes.
        real_estate_taxes = add(tax_unit, period, ["real_estate_taxes"])

        # Calculate uncapped (potential) itemized deductions.
        deduction_no_med = deduction - medical_expenses + real_estate_taxes

        # Calculate capped itemized deductions.
        capped_deduction_no_med = min_(deduction_no_med, p.cap)

        # Add back medical and dental expenses (which are not capped).
        return capped_deduction_no_med + medical_expenses



from policyengine_us.model_api import *


class labor_supply_behavioral_response(Variable):
    value_type = float
    entity = Person
    label = "earnings-related labor supply change"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        p = parameters(period).gov.simulation.labor_supply_responses
        simulation = person.simulation
        if simulation.baseline is None:
            return 0  # No reform, no impact
        if p.elasticities.income == 0 and p.elasticities.substitution.all == 0:
            return 0

        measurement_branch = simulation.get_branch(
            "lsr_measurement", clone_system=True
        )  # A branch without LSRs
        baseline_branch = simulation.get_branch("baseline").get_branch(
            "baseline_lsr_measurement", clone_system=True
        )  # Already created by default
        baseline_branch.tax_benefit_system.parameters.simulation = (
            measurement_branch.tax_benefit_system.parameters.simulation
        )

        # (system with LSRs) <- (system without LSRs used to calculate LSRs)
        #                      |
        #                      * -(baseline system without LSRs used to calculate LSRs)

        for branch in [measurement_branch, baseline_branch]:
            branch.tax_benefit_system.neutralize_variable(
                "employment_income_behavioral_response"
            )
            branch.tax_benefit_system.neutralize_variable(
                "self_employment_income_behavioral_response"
            )
            branch.set_input(
                "employment_income_before_lsr",
                period,
                person("employment_income_before_lsr", period),
            )
            branch.set_input(
                "self_employment_income_before_lsr",
                period,
                person("self_employment_income_before_lsr", period),
            )

        response = add(
            person,
            period,
            [
                "income_elasticity_lsr",
                "substitution_elasticity_lsr",
            ],
        )
        simulation = person.simulation
        del simulation.branches["baseline"].branches[
            "baseline_lsr_measurement"
        ]
        del simulation.branches["lsr_measurement"]

        simulation.macro_cache_read = False
        simulation.macro_cache_write = False

        return response



from policyengine_us.model_api import *


class snap_self_employment_income_after_expense_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "Self-employment income after the SNAP self-employment expense deduction"
    unit = USD
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        self_employment_income = add(
            spm_unit, period, ["self_employment_income"]
        )
        expense_deduction = spm_unit(
            "snap_self_employment_expense_deduction", period
        )
        return max_(self_employment_income - expense_deduction, 0)



from policyengine_us.model_api import *


class ca_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "California nonrefundable income tax credits"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/Search/Home/Confirmation"

    adds = "gov.states.ca.tax.income.credits.nonrefundable"



from policyengine_us.model_api import *


class sc_gross_earned_income(Variable):
    value_type = float
    entity = Person
    label = "South Carolina gross earned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.SC

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.sc.tax.income.credits.two_wage_earner.earned_income
        # Based on the legal code and the tax form worksheet,
        # we use earned income as the income base for the credit.
        earned_income = person("earned_income", period)
        subtractions = add(person, period, p.subtractions)
        return max_(0, earned_income - subtractions)



from policyengine_us.model_api import *


class oh_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio refundable credits"
    reference = (
        # 2021 Ohio Schedule of Credits
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/sch-cre.pdf",
        # 2022 Ohio Schedule of Credits
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/itschedule-credits.pdf",
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OH



from policyengine_us.model_api import *


class meets_wic_income_test(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Meets income test for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "Meets WIC income test"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#d_2_A_i"

    def formula(spm_unit, period, parameters):
        # Must be below the school meal reduced price tier.
        # We can't use school_meal_tier, because pregnant women have an
        # additional person for defining the poverty line
        # (for WIC but not school meals).
        income = spm_unit("school_meal_countable_income", period)
        adj_fpg = spm_unit("wic_fpg", period)
        limit = parameters(period).gov.usda.school_meals.income.limit.REDUCED
        income_fpg_ratio = income / adj_fpg
        return income_fpg_ratio <= limit



from policyengine_us.model_api import *


class ma_liheap_state_median_income_threshold(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts LIHEAP state median income threshold"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/info-details/learn-about-home-energy-assistance-heap"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.hhs.liheap
        state_median_income = spm_unit("hhs_smi", period)
        return state_median_income * p.smi_limit



from policyengine_us.model_api import *


class ca_child_care_subsidies(Variable):
    value_type = float
    entity = SPMUnit
    label = "California child care subsidies"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA
    adds = ["ca_calworks_child_care"]



from policyengine_us.model_api import *
from policyengine_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_countable_income(Variable):
    value_type = float
    entity = Person
    label = "SSI countable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382a#b"
    documentation = """
    Calculates total countable income for SSI:
      - Earned (after ignoring blind/disabled student exclusion)
      - Unearned
      - Parental deemed if child
      - Spousal deemed if married to an ineligible spouse
      - Applies standard SSI exclusions.
    """

    def formula(person, period, parameters):
        # The individual's earned income, after the blind/disabled student exclusion
        pre_reduction_earned_income = person(
            "ssi_marital_earned_income", period
        )
        blind_disabled_working_student_income = person(
            "ssi_blind_or_disabled_working_student_exclusion", period
        )
        earned_income = max_(
            pre_reduction_earned_income
            - blind_disabled_working_student_income,
            0,
        )

        # The individual's unearned + any unearned deemed from a parent
        # When both spouses are SSI eligible, ssi_marital_unearned_income and ssi_marital_earned_income
        # will already have attributed the combined earned/unearned income to each person in the
        # marital unit, so we don't need additional head/spouse checks here.
        # This occurs via the both_eligible check in those variables.
        unearned_income = person("ssi_marital_unearned_income", period)
        parent_deemed = person(
            "ssi_unearned_income_deemed_from_ineligible_parent", period
        )
        total_unearned = unearned_income + parent_deemed

        # Apply standard SSI exclusions to the individual's own income
        personal_countable = _apply_ssi_exclusions(
            earned_income,
            total_unearned,
            parameters,
            period,
        )

        # Check if both spouses are eligible
        both_eligible = person("ssi_marital_both_eligible", period)

        # Add spousal deemed income (only relevant when one spouse is ineligible)
        spousal_deemed = person(
            "ssi_income_deemed_from_ineligible_spouse", period
        )

        # Check if person is eligible (either as individual OR as eligible spouse)
        is_eligible_individual = person("is_ssi_eligible_individual", period)
        is_eligible_spouse = person("is_ssi_eligible_spouse", period)
        is_eligible = is_eligible_individual | is_eligible_spouse

        # Calculate final countable income
        return where(
            ~is_eligible,
            0,
            where(
                both_eligible,
                personal_countable / 2,  # Each eligible spouse gets half
                personal_countable
                + spousal_deemed,  # Single or with ineligible spouse
            ),
        )



from policyengine_us.model_api import *


class nc_scca_is_school_age(Variable):
    value_type = bool
    entity = Person
    label = "North Carolina SCCA school age determination"
    definition_period = YEAR
    reference = "https://ncchildcare.ncdhhs.gov/Portals/0/documents/pdf/A/ACF-118_CCDF_FFY_2022-2024_For_North_Carolina_Amendment_1.pdf?ver=C9YfIUPAFekeBA3I1mN8aA%3d%3d#page=83"
    defined_for = StateCode.NC

    def formula(person, period, parameters):
        """
        Determines if a child is considered of school age for SCCA purposes.
        Children are considered school age if they are at or above the school age threshold.
        """
        age = person("age", period)
        p = parameters(period).gov.states.nc.ncdhhs.scca.age
        return age >= p.school



from policyengine_us.model_api import *


class ca_care_eligible(Variable):
    value_type = bool
    entity = Household
    definition_period = YEAR
    label = "Eligible for California CARE program"
    documentation = "Eligible for California Alternate Rates for Energy"
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=PUC&sectionNum=739.1"
    defined_for = StateCode.CA
    adds = ["ca_care_categorically_eligible", "ca_care_income_eligible"]



from policyengine_us.model_api import *


class la_general_relief_cash_asset_limit(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    definition_period = YEAR
    label = "Limit for the Los Angeles County General Relief cash asset requirements"
    # Person has to be a resident of LA County
    defined_for = "in_la"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        married = add(spm_unit, period, ["is_married"])
        p = parameters(
            period
        ).gov.local.ca.la.general_relief.eligibility.limit.cash
        applicant_limit = where(
            married, p.applicant.married, p.applicant.single
        )
        # Recipients of the GR have an increased cash value limit
        recipient = spm_unit("la_general_relief_recipient", period)
        return where(recipient, p.recipient, applicant_limit)



from policyengine_us.model_api import *


class ky_tuition_tax_credit_eligible(Variable):
    value_type = float
    entity = TaxUnit
    label = "Eligible for the Kentucky tuition tax credit"  # Form 8863-K
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("ky_filing_status", period)
        # Married filing separate filers are ineligible.
        return filing_status != filing_status.possible_values.SEPARATE



from policyengine_us.model_api import *


class hud_gross_rent(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD gross rent"
    unit = USD
    documentation = "Gross rent for HUD programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/982.503"

    adds = ["pre_subsidy_rent", "hud_utility_allowance"]



from policyengine_us.model_api import *


class net_capital_gains(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Net capital gains before loss limitation"
    unit = USD
    adds = ["long_term_capital_gains", "short_term_capital_gains"]



from policyengine_us.model_api import *


class foreign_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Foreign tax credit"
    definition_period = YEAR
    documentation = "Foreign tax credit from Form 1116"
    unit = USD

    def formula(tax_unit, period, parameters):
        return min_(
            tax_unit("foreign_tax_credit_potential", period),
            tax_unit("foreign_tax_credit_credit_limit", period),
        )



from policyengine_us.model_api import *


class or_wfhdc_has_qualified_individual_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Check if household has eligible individuals for Oregon Working Family Household and Dependent Care Credit"
    documentation = "Oregon Working Family Household and Dependent Care Credit household eligibility"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/schedule-or-wfhdc-inst_101-195-1_2022.pdf#pahe=1",
        "https://law.justia.com/codes/oregon/2021/volume-08/chapter-315/section-315-264/",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["or"].tax.income.credits.wfhdc

        # Check if the household has a child or disabled member other than the household head.
        # Note that a disabled spouse is a qualifying individual.
        person = tax_unit.members
        age = person("age", period)
        disabled = person("is_disabled", period)
        head = person("is_tax_unit_head", period)
        qualifying_individuals = (age <= p.child_age_limit) | (
            disabled & ~head
        )

        return tax_unit.any(qualifying_individuals)



from policyengine_us.model_api import *


class ri_standard_deduction_applicable_percentage(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island standard deduction applicable percentage"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ri.gov/sites/g/files/xkgbur541/files/2021-11/2021-tax-rate-and-worksheets.pdf"
        "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022%20Tax%20Rate%20and%20Worksheets.pdf"
    )
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ri.tax.income.deductions.standard.phase_out

        agi = tax_unit("ri_agi", period)

        excess_agi = max_(agi - p.start, 0)

        excess_agi_step = np.ceil(excess_agi / p.increment)

        return max_(1 - p.percentage * excess_agi_step, 0)



from policyengine_us.model_api import *


class mt_tuition_subtraction_person(Variable):
    value_type = float
    entity = Person
    label = "Montana tuition subtraction"
    definition_period = YEAR
    reference = "https://leg.mt.gov/bills/mca/title_0150/chapter_0300/part_0210/section_0200/0150-0300-0210-0200.html"
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.subtractions.tuition
        individual_contributions = person(
            "investment_in_529_plan_indv", period
        )
        p = parameters(period).gov.states.mt.tax.income.subtractions.tuition
        capped_contributions = min_(individual_contributions, p.cap)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return capped_contributions * head_or_spouse



from policyengine_us.model_api import *


class qualified_battery_storage_technology_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified battery storage technology expenditures"
    documentation = "Expenditures for qualified battery storage technology installed in connection with a dwelling unit located in the United States and used as a residence by the taxpayer and has the capacity not less than 3kwh."
    unit = USD
    definition_period = YEAR
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=352"



from policyengine_us.model_api import *


class wv_low_income_family_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia low-income family tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = "wv_low_income_family_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        p = parameters(
            period
        ).gov.states.wv.tax.income.credits.liftc  # low_income_family_tax_credit

        wv_agi = tax_unit("wv_low_income_family_tax_credit_agi", period)
        fpg = tax_unit("wv_low_income_family_tax_credit_fpg", period)
        # modified agi limit
        fpg_amount = p.fpg_percent[filing_status] * fpg
        # condition: wv_agi < = fpg_amount
        reduced_agi = wv_agi - fpg_amount

        credit_percentage = select(
            [
                filing_status == filing_statuses.SINGLE,
                filing_status == filing_statuses.SEPARATE,
                filing_status == filing_statuses.JOINT,
                filing_status == filing_statuses.HEAD_OF_HOUSEHOLD,
                filing_status == filing_statuses.SURVIVING_SPOUSE,
            ],
            [
                p.amount.single.calc(reduced_agi),
                p.amount.separate.calc(reduced_agi),
                p.amount.joint.calc(reduced_agi),
                p.amount.head_of_household.calc(reduced_agi),
                p.amount.surviving_spouse.calc(reduced_agi),
            ],
        )

        tax_before_credits = tax_unit(
            "wv_income_tax_before_non_refundable_credits", period
        )
        return tax_before_credits * credit_percentage



from policyengine_us.model_api import *


class dc_gac_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC General Assistance for Children (GAC) countable income"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.05a#(e)"
    )
    defined_for = StateCode.DC

    adds = [
        "dc_gac_earned_income_after_disregard_person",
        "dc_gac_countable_unearned_income_person",
    ]



from policyengine_us.model_api import *


class dwks13(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "IRS Form 1040 Schedule D worksheet (part 4 of 6)"
    unit = USD

    def formula(tax_unit, period, parameters):
        e24515 = add(tax_unit, period, ["unrecaptured_section_1250_gain"])
        dwks11 = e24515 + add(
            tax_unit, period, ["capital_gains_28_percent_rate_gain"]
        )  # Sch D lines 18 and 19, respectively
        dwks09 = tax_unit("dwks09", period)
        dwks12 = min_(dwks09, dwks11)
        dwks10 = tax_unit("dwks10", period)
        return (dwks10 - dwks12) * tax_unit("has_qdiv_or_ltcg", period)



from policyengine_us.model_api import *


class la_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana standard deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.la.tax.income.deductions.standard
        if p.applies:
            filing_status = tax_unit("filing_status", period)
            return p.amount[filing_status]
        return 0



from policyengine_us.model_api import *


class wi_homestead_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin homestead credit income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleH.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleH.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wi.tax.income.credits
        income = add(tax_unit, period, p.homestead.income.sources)
        dependents = tax_unit("tax_unit_dependents", period)
        return income - dependents * p.homestead.income.exemption



from policyengine_us.model_api import *


class co_ccap_smi(Variable):
    value_type = float
    entity = SPMUnit
    label = "State median income for Colorado CCAP"
    unit = USD
    documentation = "The state median income used to determine eligibility for Colorado's Child Care Assistance Program. This differs from the HHS definition by basing it on the prior year if before October, and dividing by 12."
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        size = spm_unit("spm_unit_size", period.this_year)
        state_code = spm_unit.household("state_code_str", period.this_year)
        year = period.start.year
        month = period.start.month
        if month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        four_person_smi = parameters(instant_str).gov.hhs.smi.amount[
            state_code
        ]
        adjustment_mapping = parameters(
            instant_str
        ).gov.hhs.smi.household_size_adjustment
        first_person_rate = adjustment_mapping.first_person
        second_to_sixth_additional_rate = (
            adjustment_mapping.second_to_sixth_person
        )
        seven_or_more_additional_rate = adjustment_mapping.additional_person
        size_adjustment = (
            first_person_rate
            + second_to_sixth_additional_rate * (min_(size, 6) - 1)
            + seven_or_more_additional_rate * max_(size - 6, 0)
        )
        return four_person_smi * size_adjustment / MONTHS_IN_YEAR



from policyengine_us.model_api import *


class ca_in_home_supportive_services(Variable):
    value_type = float
    entity = Person
    label = "California In-Home Supportive Services"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://www.cdss.ca.gov/in-home-supportive-services"



from policyengine_us.model_api import *


class dc_ccsp_income_test_waived(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Income test exemption under DC Child Care Subsidy Program (CCSP)"
    definition_period = MONTH
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=11"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.age_threshold
        person = spm_unit.members
        age = person("monthly_age", period)
        is_disabled = person("is_disabled", period)
        # parent is disabled # income waived
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        has_disabled_parent = spm_unit.any(is_head_or_spouse & is_disabled)
        # child is homeless # income waived
        is_homeless = spm_unit.household("is_homeless", period)
        # parent is teen parent (age <= 19) # income waived
        is_teen_parent = age <= p.teen_parent
        has_teen_parent = spm_unit.any(is_head_or_spouse & is_teen_parent)
        return has_disabled_parent | is_homeless | has_teen_parent



from policyengine_us.model_api import *


class ca_use_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "CA Use Tax"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=22"
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        income = tax_unit("ca_agi", period)
        p = parameters(period).gov.states.ca.tax.income.use_tax
        # Compute main amount, a dollar amount based on CA AGI.
        main_amount = p.main.calc(income)
        # Switches to a percentage of income above the top main threshold.
        additional_amount = p.additional.calc(income) * income
        return main_amount + additional_amount



from policyengine_us.model_api import *


class TexasCCSChildAgeCategory(Enum):
    INFANT = "Infant"
    TODDLER = "Toddler"
    PRESCHOOL = "Preschool"
    SCHOOL_AGE = "School Age"


class tx_ccs_child_age_category(Variable):
    value_type = Enum
    entity = Person
    possible_values = TexasCCSChildAgeCategory
    default_value = TexasCCSChildAgeCategory.INFANT
    definition_period = MONTH
    defined_for = StateCode.TX
    label = "Texas Child Care Services (CCS) child age category"
    reference = (
        "https://regulations.justia.com/states/texas/title-40/part-20/chapter-809/subchapter-b/section-809-20/",
        "https://www.twc.texas.gov/programs/child-care-services",
    )

    def formula(person, period, parameters):
        # Get child's age in years (convert to months for infant threshold)
        age_years = person("age", period.this_year)
        age_months = age_years * MONTHS_IN_YEAR
        p = parameters(period).gov.states.tx.twc.ccs.age_threshold
        return select(
            [
                age_months < p.infant,  # Under 18 months = Infant
                age_years < p.toddler,  # Under 3 years = Toddler
                age_years < p.school_age,  # Under 6 years = Preschool
            ],
            [
                TexasCCSChildAgeCategory.INFANT,
                TexasCCSChildAgeCategory.TODDLER,
                TexasCCSChildAgeCategory.PRESCHOOL,
            ],
            default=TexasCCSChildAgeCategory.SCHOOL_AGE,
        )



from policyengine_us.model_api import *


class pell_grant_months_in_school(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant months of year student is in school"
    definition_period = YEAR



from policyengine_us.model_api import *


class ny_drive_clean_purchased_qualifying_vehicle(Variable):
    value_type = bool
    entity = Household
    label = "Purchased a qualifying new vehicle at an authorized dealership for the New York Drive Clean rebate program"
    definition_period = YEAR
    reference = "https://www.nyserda.ny.gov/-/media/Project/Nyserda/Files/Programs/Drive-Clean-NY/implementation-manual.pdf#pages=6,7"
    defined_for = StateCode.NY



from policyengine_us.model_api import *


class is_hispanic(Variable):
    value_type = bool
    entity = Person
    label = "hispanic"
    documentation = "Whether the person is Hispanic"
    definition_period = YEAR



from policyengine_us.model_api import *


class wv_social_security_benefits_subtraction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = (
        "Eligible for the West Virginia social security benefits subtraction"
    )
    definition_period = YEAR
    reference = (
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 33
        "https://tax.wv.gov/Documents/TaxForms/2020/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 32
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2022 LINE 32
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=25",
        # Code of West Virginia §11-21-12 (c)(8)(D)
        "https://code.wvlegislature.gov/11-21-12/",
    )
    defined_for = StateCode.WV

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.social_security_benefits
        return agi <= p.income_limit[filing_status]



from policyengine_us.model_api import *


class tx_ccs_maximum_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas Child Care Services program maximum payment"
    unit = USD
    definition_period = MONTH
    defined_for = "tx_ccs_eligible"

    def formula(spm_unit, period, parameters):
        total_reimbursement = add(spm_unit, period, ["tx_ccs_payment_rate"])
        parent_fee = spm_unit("tx_ccs_copay", period)
        return max_(total_reimbursement - parent_fee, 0)



from policyengine_us.model_api import *


class mi_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/Schedule-1.pdf",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040.pdf",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        # Michigan computes taxable income from federal AGI, additions, subtractions,
        # and exemptions
        agi = tax_unit("adjusted_gross_income", period)
        additions = tax_unit("mi_additions", period)
        subtractions = tax_unit("mi_subtractions", period)
        exemptions = tax_unit("mi_exemptions", period)

        return max_(agi + additions - subtractions - exemptions, 0)



from policyengine_us.model_api import *


class market_income(Variable):
    value_type = float
    entity = Person
    label = "Market income"
    unit = USD
    documentation = "Income from all non-government sources"
    definition_period = YEAR

    def formula(person, period, parameters):
        COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "pension_income",
            "dividend_income",
            "interest_income",
            "gi_cash_assistance",
            "capital_gains",
            "rental_income",
            "illicit_income",
            "farm_income",
            "miscellaneous_income",
            "retirement_distributions",
        ]
        return add(person, period, COMPONENTS)



from policyengine_us.model_api import *


class ma_mbta_enrolled_in_applicable_programs(Variable):
    value_type = bool
    entity = Person
    label = "Enrolled in applicable programs to receive the Massachusetts Bay Transportation Authority income-eligible reduced fare program"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mbta.com/fares/reduced/income-eligible"

    adds = "gov.states.ma.dot.mbta.income_eligible_reduced_fares.applicable_programs"



from policyengine_us.model_api import *


class va_agi_person(Variable):
    value_type = float
    entity = Person
    label = "Virginia adjusted gross income for each person"
    unit = USD
    definition_period = YEAR
    # Virginia includes an individual level definition for AGI
    reference = "https://www.tax.virginia.gov/laws-rules-decisions/rulings-tax-commissioner/13-5"
    defined_for = StateCode.VA

    def formula(person, period, parameters):
        total_agi = person.tax_unit("va_agi", period)
        person_agi = person("adjusted_gross_income_person", period)
        total_federal_agi = person.tax_unit.sum(person_agi)

        prorate = np.zeros_like(total_agi)
        mask = total_federal_agi > 0
        prorate[mask] = person_agi[mask] / total_federal_agi[mask]
        return total_agi * prorate



from policyengine_us.model_api import *


class ma_tafdc_dependent_care_deduction(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) dependent care deduction"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-275"  # (A)
    defined_for = StateCode.MA

    adds = ["ma_tafdc_dependent_care_deduction_person"]



from policyengine_us.model_api import *


class is_ccdf_continuous_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Continuous income eligibility for CCDF"



from policyengine_us.model_api import *


class mn_amt_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota alternative minimum tax (AMT) taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1mt_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1mt_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        fagi = tax_unit("adjusted_gross_income", period)
        itemizing = tax_unit("mn_itemizing", period)
        SOME_DEDUCTIONS = [
            "charitable_deduction",
            "medical_expense_deduction",
            "casualty_loss_deduction",
        ]
        some_itm_deds = itemizing * add(tax_unit, period, SOME_DEDUCTIONS)
        AMT_SUBTRACTIONS = [
            "us_govt_interest",
            "mn_charity_subtraction",
            "mn_social_security_subtraction",
        ]
        amt_subs = add(tax_unit, period, AMT_SUBTRACTIONS)
        return fagi - some_itm_deds - amt_subs



from policyengine_us.model_api import *


class MassachusettsLIHEAPHeatingType(Enum):
    HEATING_OIL_AND_PROPANE = "Heating oil and Propane"
    NATURAL_GAS = "Natural Gas"
    KEROSENE = "Kerosene"
    ELECTRICITY = "Electricity"
    OTHER = "Other"
    NONE = "None"


class ma_liheap_heating_type(Variable):
    value_type = Enum
    entity = SPMUnit
    possible_values = MassachusettsLIHEAPHeatingType
    default_value = MassachusettsLIHEAPHeatingType.ELECTRICITY
    label = "Massachusetts LIHEAP household's heating type"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/fy-2025-heap-income-eligibility-benefit-chart-may-8-2025/download"



from policyengine_us.model_api import *


class nj_total_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey total exemption allowance"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ

    adds = [
        "nj_regular_exemption",
        "nj_senior_exemption",
        "nj_dependents_exemption",
        "nj_blind_or_disabled_exemption",
        "nj_dependents_attending_college_exemption",
    ]



from policyengine_us.model_api import *


class tx_tanf_recognizable_needs_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Texas TANF recognizable needs test"
    definition_period = MONTH
    reference = "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Recognizable needs test applies to continuing recipients
        # Household passes if recognizable needs > countable income

        budgetary_needs = spm_unit("tx_tanf_budgetary_needs", period)
        countable_income = spm_unit("tx_tanf_countable_income", period)
        p = parameters(period).gov.states.tx.tanf.needs_standard

        # Recognizable needs is 25% of budgetary needs
        recognizable_needs = budgetary_needs * p.recognizable_needs.rate

        return countable_income < recognizable_needs



from policyengine_us.model_api import *


class energy_efficient_home_improvement_credit_potential(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Potential value of the Energy efficient home improvement credit"
    documentation = "Residential clean energy credit"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/25C"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.irs.credits.energy_efficient_home_improvement
        if not p.in_effect:
            return 0

        heat_pump_etc = tax_unit(
            "capped_heat_pump_heat_pump_water_heater_biomass_stove_boiler_credit",
            period,
        )
        total = add(tax_unit, period, p.qualified_expenditures.credits)
        # Cap the total.
        capped_total = min_(total, p.cap.annual.total)
        # Before the lifetime limit, it can either be the total or the heat pump/etc.
        pre_lifetime_limit = max_(capped_total, heat_pump_etc)
        # Apply lifetime limitation.
        prior_credits = tax_unit(
            "prior_energy_efficient_home_improvement_credits", period
        )
        remaining_credit = p.cap.lifetime.total - prior_credits
        return min_(remaining_credit, pre_lifetime_limit)
